#!/usr/bin/env python
import numpy as np
from scipy.optimize import fmin
from pyjams.const import R, T0
from pyjams.air_humidity import esat
from pyjams.jams.errormeasures import mae
from pyjams.sce import sce
'''
    leafmodel: contains routines to model photosynthesis and stomatal conductance
    of canopies or leaves.


    major functions contained:
    --------------------------
    opt_leafmodel:  optimization against observed transpiration and
                    gross primary production
    cost_leafmodel: cost function for optimization
    leafmodel:      model for photosynthesis and stomatal conductance
    twoleafmodel:   uses leafmodel in a sunlit and shaded scheme
    farquhar:       photosysnthesis model
    leuning:        stomatal conductance modeling
    ball_berry:     stomatal conductance modeling


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP & MC, Jul 2014
'''
###############################################################################
# Constants

# Michaelis-Menten constant for CO2 at 25 degC
Kc25    = 460.e-6 # [mol/mol]
# Activation energy for CO2
E_Kc    = 59356.  # [J]
# O2 partial pressure
O       = 0.21    # [mol/mol]
# Activation energy for O
E_O     = 35948.  # [J]
# Michaelis-Menten constant for O2 at 25 degC
Ko25    = 0.33    # [mol/mol]
# Activation energy for O2
E_Ko    = 35948.  # [J]
# effiency of photon capture
alpha   = 0.28    # [-]
# Activation energy for dark respiration Rd
E_Rd    = 50967.  # [J]
# decrease rate of Vcmax above temperature optimum (Medlyn approach)
HdV     = 200000. # [J/mol]
# decrease rate of Jmax above temperature optimum (Medlyn approach)
HdJ     = 220000. # [J/mol]
# Activation energy for Vcmax
E_Vcmax = 58520.  # [J]
# Activation energy for Jcmax
E_Jmax  = 35870.  # [J]

###############################################################################
def opt_leafmodel(params, Eobs, GPPobs, ci_ini, Tl, PAR, ea, ca, ga, gb, P,
                  PARdiff=None, fsun=None, dooptim=False, loc_search=True,
                  silent=True, n=5, eta=0.9):
    '''
    Optimization routine for leafmodel or twoleafmodel to infer model parameters
    fitting gross primary productivity and transpiration to observed values.


    Definition
    ----------
    def opt_leafmodel(params, Eobs, GPPobs, ci_ini, Tl, PAR, ea, ca, ga, gb, P,
                  PARdiff=None, fsun=None, dooptim=False, loc_search=True,
                  n=5, eta=0.9):


    Input
    -----
    params        dict, parameters for leafmodel. Parameters given here decide
                  which conductance model and temperature dependence model is
                  used and which parameters shall be optimized and which not.
                  params CAN contain following keys with its respective initial
                  values: 'm','b','d0','Vcmax25','delSV','delSJ','Topt','omega'.
                  All parameters given in this way will be optimized. You can
                  give parameters also with the ending 'fix' like 'mfix' or
                  Toptfix', then they will not be optimized but kept fix during
                  optimization.
                  params MUST contain 'm' and 'b' (or with the 'fix' ending).
                  If you give 'd0' or 'd0fix', the Leuning model will be used,
                  else the Ball&Berry model will be used.
                  params MUST contain 'Vcmax25' or 'Vcamx25fix'. If you give
                  'Topt' and 'omega' (or with the fix ending), the June model
                  is used. If you give 'delSV' and 'delSJ' (or with the fix
                  ending), the Medlyn model is used. If you only give 'delSJ' or
                  'delSJfix', the von Caemmerer model is used. For model
                  descriptions see leafmodel.
    Eobs          np.array(N), observed transpiration [mol/m2s]
    GPPobs        np.array(N), observed gross primary productivity [mol/m2s]
    ci_ini, Tl, PAR, ea, ca, ga, gb, P: same as for leafmodel


    Optional Input
    --------------
    PARdiff, fsun if both set to None (default) leafmodel is used, else
                  twoleafmodel is used and input is like twoleafmodel
    dooptim       bool, if True optimization is conducted, is False (default)
                  no optimization is conducted and output is calculated using
                  the input params
    loc_search    bool, if True optimization uses Nelder-Mead local search
                  (scipy.optimize.fmin) (default), if False shuffled complex
                  evolution global (jams.sce) search is used. fmin is faster, but
                  params must be close to the optimum parameters. sce is slower
                  but always finds the minimum, even if params is far away from
                  optimum. The parameter limits of sce are set to 0 and 2*params
                  for each parameter.
    silent        bool, if True (default), no output during optimization, if
                  False objective function value is printed to the console at
                  each cost function evaluation
    n, eta        same as for leafmodel


    Output
    ------
    opt_params    list, optimized parameters (does not contain the fixed values)
    ci_conv, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod:
                  same as leafmodel


    Examples
    --------
    >>> # Define input
    >>> ci_ini = np.array([3.361e-4, 3.354e-4, 3.350e-4, 3.345e-4, 3.337e-4])
    >>> Tl     = np.array([21.041, 21.939, 21.997, 21.026, 20.237])
    >>> PAR    = np.array([1.144e-3, 1.294e-3, 1.445e-3, 1.733e-3, 1.752e-3])
    >>> ea     = np.array([1636.967, 1609.575, 1569.418, 1561.334, 1460.839])
    >>> ca     = np.array([4.202e-4, 4.193e-4, 4.188e-4, 4.182e-4, 4.171e-4])
    >>> ga     = np.array([3.925, 5.998, 3.186, 7.924, 5.501])
    >>> gb     = np.array([1.403, 1.678, 1.732, 1.691, 1.782])
    >>> P      = np.array([99893.7, 99907.7, 99926.5, 99928.0, 99924.7])
    >>> PARdiff= np.array([3.746e-4, 4.059e-4, 4.428e-4, 4.663e-4, 4.849e-4])
    >>> fsun   = np.array([0.682, 0.703, 0.718, 0.729, 0.737])
    >>> Eobs   = np.array([3.552e-3, 3.854e-3, 3.828e-3, 4.455e-3, 5.322e-3])
    >>> GPPobs = np.array([1.293e-5, 1.351e-5, 1.339e-5, 1.318e-5, 1.139e-5])

    >>> # use default methods Leuning + June with fixed b, Topt and omega
    >>> params = {'m':15., 'bfix':0.0041, 'd0':5.e3,\
    'Vcmax25':60.e-6, 'Toptfix':25., 'omegafix':18.}
    >>> opt_params, ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod = \
    opt_leafmodel(params, Eobs, GPPobs, ci_ini, Tl, PAR, ea, ca, ga, gb, P,\
    PARdiff, fsun, dooptim=True)
    >>> # print GPP [mumol(m2s] and E [mmol/m2s]
    >>> print(np.ma.round(GPPmod*1.e6, 3))
    [12.877 13.29 13.384 13.273 13.013]
    >>> print(np.ma.round(Emod*1.e3, 3))
    [3.546 4.442 4.306 4.31 4.157]
    >>> # print optimized parameters
    >>> m, d0, Vcmax25 = opt_params
    >>> print(np.round(m, 3))
    48.346
    >>> print(np.round(d0/100., 3)) #[hPa]
    4.343
    >>> print(np.round(Vcmax25*1.e6, 3)) #[mumol/m2s]
    65.763


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP+MC, Jul 2014
    '''
    # select methods according to given paramters and sort them according to
    # fixed or non fixed
    gs_method = 'BB'
    fix       = {}

    if ('m' in params):
        guess = [params['m']]
    else:
        fix.update({'mfix':params['mfix']})
    if ('b' in params):
        guess += [params['b']]
    else:
        fix.update({'bfix':params['bfix']})

    if ('d0' in params):
        guess += [params['d0']]
        gs_method = 'Leuning'
    elif('d0fix' in params):
        fix.update({'d0fix':params['d0fix']})
        gs_method = 'Leuning'

    if ('Vcmax25' in params):
        guess += [params['Vcmax25']]
    else:
        fix.update({'Vcmax25fix':params['Vcmax25fix']})

    temp_method = 'June'
    if ('Topt' in params):
        guess += [params['Topt']]
        assert ('omega' in params) or ('omegafix' in params)
    elif ('Toptfix' in params):
        fix.update({'Toptfix':params['Toptfix']})
        assert ('omega' in params) or ('omegafix' in params)
    if ('omega' in params):
        guess += [params['omega']]
    elif ('omegafix' in params):
        fix.update({'omegafix':params['omegafix']})
    else:
        if ('delSV' in params):
            guess += [params['delSV']]
            temp_method = 'Medlyn'
        elif ('delSVfix' in params):
            fix.update({'delSVfix':params['delSVfix']})
            temp_method = 'Medlyn'
        else:
            temp_method = 'Caemmerer'

        if ('delSJ' in params):
            guess += [params['delSJ']]
        elif ('delSJfix' in params):
            fix.update({'delSJfix':params['delSJfix']})

    # optimization
    if dooptim:
        # use local nelmin optimization
        if loc_search:
            opt_params = fmin(cost_leafmodel, guess,
                              args=(Eobs, GPPobs, ci_ini, Tl, PAR, ea, ca, ga,
                                    gb, P, PARdiff, fsun, n, eta, gs_method,
                                    temp_method, fix, silent), disp=0)
        # use global sce optimization
        else:
            def wrap_opt(opt_params):
                obj = cost_leafmodel(opt_params, Eobs, GPPobs, ci_ini, Tl, PAR,
                                     ea, ca, ga, gb, P, PARdiff,fsun, n, eta,
                                     gs_method, temp_method, fix, silent)
                return obj
            guess=np.array(guess)
            opt_params = sce(wrap_opt, guess, np.zeros_like(guess) , guess*2.,
                             maxn=10000, kstop=15)
    else:
        opt_params = guess

    # distribute parameters
    par1, par2 = dist_params(opt_params, fix, gs_method, temp_method)

    # calculate output
    if (PARdiff is None) and (fsun is None):
        ci, gcmol_c, gcmol_h, gsmol_c, gsmol_h, no_conv, Jc, Je, Rd, GPPmod, Emod =\
            leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2,
                      gs_method=gs_method, temp_method=temp_method, n=n, eta=eta)
    elif (PARdiff is not None) and (fsun is not None):
        ci, gcmol_c, gcmol_h, gsmol_c, gsmol_h, no_conv, Jc, Je, Rd, GPPmod, Emod =\
            twoleafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, PARdiff, fsun, par1, par2,
                         gs_method=gs_method, temp_method=temp_method, n=n, eta=eta)
    else:
        raise ValueError('opt_leafmodel: if two leaf model is desired, PARdiff and fsun must be given')

    return opt_params, ci, gcmol_c, gcmol_h, gsmol_c, gsmol_h, no_conv, Jc, Je, Rd, GPPmod, Emod

###############################################################################
def cost_leafmodel(params, Eobs, GPPobs, ci_ini, Tl, PAR, ea, ca, ga, gb, P,
                   PARdiff, fsun, n, eta, gs_method, temp_method, fix,
                   silent=True):
    '''
    opt_leafmodel: cost function for optimization
    (input see leafmodel and twoleafmodel)

    returns: obj objective function value
    '''
    # distribute parameters
    par1, par2 = dist_params(params, fix, gs_method, temp_method)

    if (PARdiff is None) and (fsun is None):
        # run the model with current parameters
        ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod =\
            leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2,
                      gs_method=gs_method, temp_method=temp_method, n=n,
                      eta=eta)
    elif (PARdiff is not None) and (fsun is not None):
        ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod =\
            twoleafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, PARdiff, fsun,
                         par1, par2, gs_method=gs_method,
                         temp_method=temp_method, n=n, eta=eta)
    else:
        raise ValueError('cost_leafmodel: if two leaf model is desired, PARdiff and fsun must be given')

    # mask values where ci did not converged
    Emod.mask[no_conv], GPPmod.mask[no_conv] = True, True
    # calculate range of observations
    Eobs_r   = np.ma.ptp(Eobs)
    GPPobs_r = np.ma.ptp(GPPobs)
    # if for some time steps ci did not converge, tell fmin that's bad!
    if no_conv.size>0:
        return 1000.
    # else calculate objective function
    else:
        # combined mean absolute error
        obj = (
               mae(Eobs/Eobs_r, Emod/Eobs_r)**6 +
               mae(GPPobs/GPPobs_r, GPPmod/GPPobs_r)**6
               )**(1./6.)
        if not silent:
            print(obj)

        return obj

###############################################################################
def dist_params(params, fix, gs_method, temp_method):
    '''
    opt_leafmodel+cost_leafmodel: distibutes parameters to variables according
    to selected methods
    '''
    params = np.insert(params, 0, fix['mfix']) if 'mfix' in fix else params
    params = np.insert(params, 1, fix['bfix']) if 'bfix' in fix else params

    if gs_method=='Leuning':
        params = np.insert(params, 2, fix['d0fix']) if 'd0fix' in fix else params
        params = np.insert(params, 3, fix['Vcmaxfix']) if 'Vcmaxfix' in fix else params
        if temp_method=='June':
            params = np.insert(params, 4, fix['Toptfix']) if 'Toptfix' in fix else params
            params = np.insert(params, 5, fix['omegafix']) if 'omegafix' in fix else params
        elif temp_method=='Medlyn':
            params = np.insert(params, 4, fix['delSVfix']) if 'delSVfix' in fix else params
            params = np.insert(params, 5, fix['delSJfix']) if 'delSJfix' in fix else params
        elif temp_method=='Caemmerer':
            params = np.insert(params, 4, fix['delSJfix']) if 'delSJfix' in fix else params
    else:
        params = np.insert(params, 2, fix['Vcmaxfix']) if 'Vcmaxfix' in fix else params
        if temp_method=='June':
            params = np.insert(params, 3, fix['Toptfix']) if 'Toptfix' in fix else params
            params = np.insert(params, 4, fix['omegafix']) if 'omegafix' in fix else params
        elif temp_method=='Medlyn':
            params = np.insert(params, 3, fix['delSVfix']) if 'delSVfix' in fix else params
            params = np.insert(params, 4, fix['delSJfix']) if 'delSJfix' in fix else params
        elif temp_method=='Caemmerer':
            params = np.insert(params, 3, fix['delSJfix']) if 'delSJfix' in fix else params

    if gs_method=='Leuning':
        par1 = params[:3]
        par2 = params[3:]
    if gs_method=='BB':
        par1 = params[:2]
        par2 = params[2:]

    return par1, par2

###############################################################################
def leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2,
              gs_method='Leuning', temp_method='June', n=5, eta=0.9):
    '''
    Model to calculate photosynthesis and stomatal conductance of canopies. The
    Farquhar model (Farquhar et al, 1980) is used for photosynthesis. Stomatal
    conductance can either be modeled with Ball & Berry (1987) or with
    Leuning (1995). Aerodynamic and leaf boundary layer conductances are used
    to calculate leaf surface concentrations and the model iterates until a
    conversion of leaf internal CO2 concentration. Canopy transpiration and
    gross primary production are among the results. The temperature dependency
    of the photosynthesis model can be modeled according to June et al. (2004),
    Medlyn et al. (2002) and von Caemmerer (2000).


    Definition
    ----------
    def leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2,
              gs_method='Leuning', temp_method='June', n=5, eta=0.9):


    Input
    -----
    ci_ini      np.array(N), initial leaf internal co2 concentration [mol/mol]
                Can be approximated with 0.8*ca
    Tl          np.array(N), leaf temperature [degC]
    PAR         np.array(N), total (direct+diffuse) photon flux density
                [mol/m2s]
    ea          np.array(N), vapour pressure of the atmosphere [Pa]
    ca          np.array(N), atmospheric CO2 concentration [mol/mol]
    ga          np.array(N), aerodynamic conductivity [mol/m2s] (of heat and
                water, carbon is calulated in the model)
    gb          np.array(N), leaf boundary layer conductivity [mol/m2s] (of heat
                and water, carbon is calulated in the model)
    P           np.array(N), atmospheric pressure [Pa]
    par1        if gs_method='Leuning' (default): list(3)
                    [0] m    slope of the Leuning model [mol(H2O)/mol(air)]
                    [1] b    offset of the Leuning model [mol(H2O)/m2leaf s]
                    [2] d0   sensitivity parameter to vpds [Pa]
                if gs_method='BB': list(2)
                    [0] m    slope of the BB model [mol(H2O)/mol(air)]
                    [1] b    offset of the BB model [mol(H2O)/m2leaf s]
    par2        if temp_method is June (default): list(3)
                    [0] Vcmax25 maximum carboxylation rate at 25 degC [mol/m2s]
                    [1] Topt    optimum temperature of Jmax [degC]
                    [2] omega   the difference in temperature from Topt at
                                which Jmax falls to e-1 (0.37) of its value
                                at Topt. (In the paper its on average
                                18+-0.6 degC)
                if temp_method is Medlyn: list(3)
                    [0] Vcmax25 maximum carboxylation rate at 25 degC [mol/m2s]
                    [1] delSV   entropy factor for Vcmax25 [J/mol/K]
                    [2] delSJ   entropy factor for Jmax25 [J/mol/K]
                if temp_method is Caemmerer: list(2)
                    [0] Vcmax25 maximum carboxylation rate at 25 degC [mol/m2s]
                    [1] delSJ   entropy factor for Jmax25 [J/mol/K]


    Optional Input
    --------------
    gs_method   str, stomatal conductance method: 'Leuning' (default), 'BB'
                (Ball&Berry)
    temp_method str, temperature dependecy method: 'June', 'Medlyn', 'Caemmerer'
    n           int, number of iteration for leaf internal CO2 convergence
    eta         float, smoothing factor for transition between Jmax and Vcmax
                dominated assimilation rate [-]


    Output
    ------
    ci_conv     np.array(N), leaf internal CO2 concentration [mol/mol]
    gc_c        np.array(N), canopy conductivity for carbon [mol/m2s]
    gc_h        np.array(N), canopy conductivity for water and heat [mol/m2s]
    gs_c        np.array(N), stomatal conductivity for carbon [mol/m2s]
    gs_h        np.array(N), stomatal conductivity for water and heat [mol/m2s]
    no_conv     np.array(x), indices of values where no ci conversion was
                achieved with n iterations
    Jc          np.array(N), carboxylation limited assimilation rate [mol/m2s]
    Je          np.array(N), electron transport limited assimilation rate
                [mol/m2s]
    Rd          np.array(N), leaf dark respiration rate [mol/m2s]
    GPPmod      np.array(N), gross primary productivity [mol/m2s]
    Emod        np.array(N), transpiration [mol/m2s]


    Examples
    --------
    >>> # Define input
    >>> ci_ini = np.array([3.361e-4, 3.354e-4, 3.350e-4, 3.345e-4, 3.337e-4])
    >>> Tl     = np.array([21.041, 21.939, 21.997, 21.026, 20.237])
    >>> PAR    = np.array([1.144e-3, 1.294e-3, 1.445e-3, 1.733e-3, 1.752e-3])
    >>> ea     = np.array([1636.967, 1609.575, 1569.418, 1561.334, 1460.839])
    >>> ca     = np.array([4.202e-4, 4.193e-4, 4.188e-4, 4.182e-4, 4.171e-4])
    >>> ga     = np.array([3.925, 5.998, 3.186, 7.924, 5.501])
    >>> gb     = np.array([1.403, 1.678, 1.732, 1.691, 1.782])
    >>> P      = np.array([99893.7, 99907.7, 99926.5, 99928.0, 99924.7])

    >>> # use default methods Leuning + June
    >>> par1   = [15., 0.0041, 5.e3]
    >>> par2   = [60.e-6, 25., 18.]
    >>> ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod = \
    leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2)
    >>> # print GPP [mumol(m2s] and E [mmol/m2s]
    >>> print(np.ma.round(GPPmod*1.e6, 3))
    [12.333 12.685 12.696 12.462 12.157]
    >>> print(np.ma.round(Emod*1.e3, 3))
    [3.339 4.338 4.287 4.02 3.845]

    >>> # use methods Ball&Berry + Medlyn
    >>> par1   = [15., 0.0041]
    >>> par2   = [60.e-6, 700., 700.]
    >>> ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod = \
    leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2,\
    gs_method='BB', temp_method='Medlyn')
    >>> # print GPP [mumol(m2s] and E [mmol/m2s]
    >>> print(np.ma.round(GPPmod*1.e6, 3))
    [14.286 14.496 14.499 14.526 14.287]
    >>> print(np.ma.round(Emod*1.e3, 3))
    [2.718 3.326 3.22 3.167 2.994]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP+MC, Jul 2014
    '''

    # array for convergence of leaf internal carbon concentration ci
    ci_conv = np.ma.empty((n,ci_ini.size))
    # initial conditions for ci convergence
    ci_conv[0,:] = ci_ini # [mol(CO2)/mol(air)]
    Emod         = 0.     # [mol(H2O)/m2leaf s]
    gs_h         = 1.     # [mol(H2O)/m2leaf s]

    # atmospheric (wa) and leaf internal (wi) water concentration
    wa = ea/P    # [mol(H2O)/mol(air)]
    ei = esat(Tl+T0)
    wi = ei/P # [mol(H2O)/mol(air)]

    # loop for ci convergence
    for i in range(n-1):
        # assimilation A, carboxylation limited assimilation Jc,
        # electron transport limited assimilation Je and leaf dark respiration
        # rate Rd, all in [mol/m2leaf s]
        A, Jc, Je, Rd = farquhar(Tl, ci_conv[i,:], PAR, par2,
                                 temp_method=temp_method, eta=eta)

        # aerodynamic + leaf boundary layer conductances
        gab_c         = 1./(1.37/gb + 1./ga) # [mol(CO2)/m2leaf s]
        gab_h         = 1./(1./gb + 1./ga) # [mol(H2O)/m2leaf s]

        # carbon concentration at the leaf surface
        cs            = ca - A/gab_c #[mol(CO2)/mol(air)]
        # water concentration at the leaf surface
        ws            = wa - Emod/gab_h # [mol(H2O)/mol(air)]
        # vapour pressure at the leaf surface
        es            = ws*P         # [Pa]
        # relative humidity at the leaf surface
        rHs           = es/ei     # [-]

        # stomatal conductance for carbon [mol(CO2)/m2leaf s]
        if gs_method=='Leuning':
            gs_c = leuning(A, ei-es, cs, Tl, par1)
            #gs_c = leuning(A, wi-ws, cs, Tl, par1)
        if gs_method=='BB':
            gs_c = ball_berry(A, rHs, cs, par1)
        # stomatal conductance for water [mol(H2O)/m2leaf s]
        gs_h = gs_c * 1.6

        # canopy conductances
        gc_c           = 1./(1./gs_c + 1./gab_c) # [mol(CO2)/m2leaf s]
        gc_h           = 1./(1./gs_h + 1./gab_h) # [mol(H2O)/m2leaf s]
        # new ci value
        ci_conv[i+1,:] = ca - A/(gc_c) # [mol(CO2)/mol(air)]
        # new modelled transpiration
        Emod           = gc_h * (wi-wa) # [mol(H2O)/m2leaf s]

    # number of non-coverging ci values
    no_conv = np.ma.where(np.ma.abs(ci_conv[-1,:]-ci_conv[-2,:])>10.e-6)[0]

    # modeled transpiration
    Emod   = gc_h * ((ei-ea)/P) # [mol(H2O)/m2leaf s]
    # modeled gross primary production
    GPPmod = gc_c * (ca - ci_conv[-1,:]) + Rd       # [mol(CO2)/m2leaf s]

    return ci_conv[-1,:], gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod

###############################################################################
def twoleafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, PARdiff, fsun, par1, par2,
                 gs_method='Leuning', temp_method='June', n=5, eta=0.9):
    '''
    Model to calculate photosynthesis and stomatal conductance of canopies in a
    two leaf scheme. It calculates leafmodel for a sunlit and shaded leaf
    fraction of the canopy. Inputs are the same as for leafmodel plus PARdiff
    and fsun.

    Definition
    ----------
    def twoleafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, PARdiff, fsun, par1, par2,
                 gs_method='Leuning', temp_method='June', n=5, eta=0.9):


    Input
    -----
    same as for leafmodel except:
    PARdiff     np.array(N), diffuse photon flux density [mol/m2s]
    fsun        np.array(N), fraction of sunlit leaf area [-]


    Optional Input
    --------------
    same as for leafmodel


    Output
    ------
    same as for leafmodel


    Examples
    --------
    >>> # Define input
    >>> ci_ini = np.array([3.361e-4, 3.354e-4, 3.350e-4, 3.345e-4, 3.337e-4])
    >>> Tl     = np.array([21.041, 21.939, 21.997, 21.026, 20.237])
    >>> PAR    = np.array([1.144e-3, 1.294e-3, 1.445e-3, 1.733e-3, 1.752e-3])
    >>> ea     = np.array([1636.967, 1609.575, 1569.418, 1561.334, 1460.839])
    >>> ca     = np.array([4.202e-4, 4.193e-4, 4.188e-4, 4.182e-4, 4.171e-4])
    >>> ga     = np.array([3.925, 5.998, 3.186, 7.924, 5.501])
    >>> gb     = np.array([1.403, 1.678, 1.732, 1.691, 1.782])
    >>> P      = np.array([99893.7, 99907.7, 99926.5, 99928.0, 99924.7])
    >>> PARdiff= np.array([3.746e-4, 4.059e-4, 4.428e-4, 4.663e-4, 4.849e-4])
    >>> fsun   = np.array([0.682, 0.703, 0.718, 0.729, 0.737])

    >>> # use default methods Leuning + June
    >>> par1   = [15., 0.0041, 5.e3]
    >>> par2   = [60.e-6, 25., 18.]
    >>> ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod = \
    twoleafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, PARdiff, fsun, par1, par2)
    >>> # print GPP [mumol(m2s] and E [mmol/m2s]
    >>> print(np.ma.round(GPPmod*1.e6, 3))
    [11.861 12.255 12.344 12.176 11.924]
    >>> print(np.ma.round(Emod*1.e3, 3))
    [3.253 4.232 4.205 3.952 3.792]

    >>> # use methods Ball&Berry + Medlyn
    >>> par1   = [15., 0.0041]
    >>> par2   = [60.e-6, 700., 700.]
    >>> ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod = \
    twoleafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, PARdiff, fsun, par1, par2,\
    gs_method='BB', temp_method='Medlyn')
    >>> # print GPP [mumol(m2s] and E [mmol/m2s]
    >>> print(np.ma.round(GPPmod*1.e6, 3))
    [13.482 13.785 13.904 14.002 13.839]
    >>> print(np.ma.round(Emod*1.e3, 3))
    [2.612 3.206 3.127 3.083 2.926]


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP+MC, Jul 2014
    '''
    # sunlit leaves
    ci_l, gc_c_l, gc_h_l, gs_c_l, gs_h_l, no_conv_l, Jc_l, Je_l, Rd_l,\
    GPP_l, E_l = leafmodel(ci_ini, Tl, PAR, ea, ca, ga, gb, P, par1, par2,
                           gs_method=gs_method, temp_method=temp_method,
                           n=n, eta=eta)

    # shaded leaves
    ci_s, gc_c_s, gc_h_s, gs_c_s, gs_h_s, no_conv_s, Jc_s, Je_s, Rd_s,\
    GPP_s, E_s = leafmodel(ci_ini, Tl, PARdiff, ea, ca, ga, gb, P, par1, par2,
                           gs_method=gs_method, temp_method=temp_method,
                           n=n, eta=eta)

    # combine
    ci      = fsun*ci_l   + (1.-fsun)*ci_s
    gc_c    = fsun*gc_c_l + (1.-fsun)*gc_c_s
    gc_h    = fsun*gc_h_l + (1.-fsun)*gc_h_s
    gs_c    = fsun*gs_c_l + (1.-fsun)*gs_c_s
    gs_h    = fsun*gs_h_l + (1.-fsun)*gs_h_s
    no_conv = np.concatenate((no_conv_l,no_conv_s))
    Jc      = fsun*Jc_l   + (1.-fsun)*Jc_s
    Je      = fsun*Je_l   + (1.-fsun)*Je_s
    Rd      = fsun*Rd_l   + (1.-fsun)*Rd_s
    GPPmod  = fsun*GPP_l  + (1.-fsun)*GPP_s
    Emod    = fsun*E_l    + (1.-fsun)*E_s

    return ci, gc_c, gc_h, gs_c, gs_h, no_conv, Jc, Je, Rd, GPPmod, Emod

###############################################################################
def farquhar(Tl, ci, PAR, par, temp_method='June', eta=0.9):
    '''
    leafmodel: photosynthesis model
    Tl : leaf temperature [degC]
    ci : leaf intercellular co2 concentration [mol/mol]
    PAR: photosynthetic active photon flux density [mol/m2leaf s]
    par: if temp_method is June:
         [0] VCmax25: maximum carboxylation rate at 25 degC [mol/m2leaf s]
         [1] Topt:    optimum temperature of Jmax [degC]
         [2] omega:   the difference in temperature from Topt at which Jmax
                      falls to e-1 (0.37) of its value at Topt. (In the paper
                      its on average 18+-0.6 degC)
         if temp_method is Medlyn:
         [0] VCmax25: maximum carboxylation rate at 25 degC [mol/m2leaf s]
         [1] delSV:   entropy factor for VCmax [J/mol/K]
         [2] delSJ:   entropy factor for Jmax [J/mol/K]
         if temp_method is Caemmerer:
         [0] VCmax25: maximum carboxylation rate at 25 degC [mol/m2leaf s]
         [1] delSJ:   entropy factor for Jmax [J/mol/K]
    temp_method: temperature dependecy methods: 'June', 'Medlyn', 'Caemmerer'
    eta: smoothiing factor for transition between Jmax and Vcmax dominated
         assimilation rate [-]

    returns:
    A : assimilation rate [mol/m2leaf s]
    Jc: carboxylation limited assimilation flux [mol/m2leaf s]
    Je: electron transport limited assimilation flux [mol/m2leaf s]
    Rd: leaf dark respiration [mol/m2leaf s]
    '''
    if temp_method=='June':
        assert len(par)==3, 'farquhar: wrong parameters for method June'
        Vcmax25 = par[0]
        Topt    = par[1]
        omega   = par[2]
    elif temp_method=='Medlyn':
        assert len(par)==3, 'farquhar: wrong parameters for method Medlyn'
        Vcmax25 = par[0]
        delSV   = par[1]
        delSJ   = par[2]
    elif temp_method=='Caemmerer':
        assert len(par)==2, 'farquhar: wrong parameters for method Caemmerer'
        Vcmax25 = par[0]
        delSJ   = par[1]
    else:
        raise ValueError('farquhar: unknown temperature method')

    # CO2 compensation point
    gamma_star = compensationpoint(Tl) # [mol/mol]
    # maximum electron transport rate at 25 degC
    Jm25 = 1.67*Vcmax25 # [mol/m-2s-1] Medlyn et al. 2002
    # leaf or dark respiration
    Rd25 = 0.011*Vcmax25 # [mol/m-2s-1]

    # calculate temperature dependencies
    Kc    = ArrheniusTemp(Kc25, E_Kc, Tl)
    Ko    = ArrheniusTemp(Ko25, E_Ko, Tl)
    Rd    = ArrheniusTemp(Rd25, E_Rd, Tl)
    if temp_method=='June':
        Vcmax = ArrheniusTemp(Vcmax25, E_Vcmax, Tl)
        Jm    = JuneTemp(Jm25, Topt, Tl, omega)
    elif temp_method=='Medlyn':
        Vcmax = MedlynTemp(Vcmax25, E_Vcmax, delSV, Tl, HdV)
        Jm    = MedlynTemp(Jm25, E_Jmax, delSJ, Tl, HdJ)
    elif temp_method=='Caemmerer':
        Vcmax = ArrheniusTemp(Vcmax25, E_Vcmax, Tl)
        Jm    = MedlynTemp(Jm25, E_Jmax, delSJ, Tl, HdJ)

    # carboxylation limited assimilation flux [mol/m2leaf s]
    Jc = Vcmax * (ci - gamma_star) / (ci + Kc*(1. + O/Ko))
    # electron transport limited assimilation flux [mol/m2leaf s]
    Je = (alpha * PAR * Jm / np.ma.sqrt(alpha**2 * PAR**2 + Jm**2)) * \
         ((ci - gamma_star) / (4.*(ci + 2. * gamma_star)))
    # assimilation [mol/m2leaf s]
    A = smoothmin(Jc, Je, eta) - Rd

    return A , Jc, Je, Rd# [mol/m2leaf s]

###############################################################################
def compensationpoint(T):
    '''
    leafmodel: CO2 compensation point
    T: temperature [degC]

    returns:
    gamma_star: CO2 compensation point [mol/mol]
    '''
    gamma_star = 1.7e-6 * T
    return gamma_star

###############################################################################
def ball_berry(A, rH, cs, par):
    '''
    leafmodel: stomatal conductance
    A: carbon assimilation rate [mol/m2leaf s]
    rH: relative humidity [-]
    cs: carbon mixing ratio at the leaf surface [mol/mol]
    par: [0] m: slope [mol(H2O)/mol(air)]
         [1] b: offset [mol(H2O)/m2leaf s]

    returns:
    gs: stomatal conductance [mol/m2leaf s]
    '''
    m = par[0]
    b = par[1]
    return m * A * rH/cs + b

###############################################################################
def leuning(A, vpds, cs, Tl, par):
    '''
    leafmodel: stomatal conductance
    A: carbon assimilation rate [mol/m2leaf s]
    vpds: vapour pressure deficit at the leaf surface [Pa]
    cs: carbon mixing ratio at the leaf surface [mol/mol]
    Tl: temperature at the leaf surface [degC]
    par: [0] m:  slope [mol(H2O)/mol(air)]
         [1] b:  offset [mol(H2O)/m2leaf s]
         [2] d0: sensitivity parameter to vpds [mol(H2O)/mol(air) or Pa]

    returns:
    gs: stomatal conductance [mol/m2leaf s]
    '''
    m  = par[0]
    b  = par[1]
    d0 = par[2]
    # CO2 compensation point
    gamma_star = compensationpoint(Tl) # [mol/mol]
    return m * A/(cs-gamma_star) * 1./(1.+vpds/d0) + b

###############################################################################
def smoothmin(x,y,eta):
    '''
    leafmodel: smooth minimum
    Smooth minumum function for the transition from Jmax to Vcmax dominated
    assimilation rate
    '''
    z = (x+y)**2 - 4.*eta*x*y
    z = np.ma.maximum(z,1.e-18)
    return (x+y-np.ma.sqrt(z)) / (2.*eta)

###############################################################################
def ArrheniusTemp(r25, E, T):
    '''
    leafmodel: temperature dependency
    Arrhenius style temperature dependency
    r25: activity rate at 25 degC [unit depends on input]
    E: activation energy [J]
    T: temperature [degC]

    returns: r activity rate at T [unit depends on input]
    '''
    return r25 * np.ma.exp(((T-25.)*E)/(298.*R*(T+273.)))

###############################################################################
def JuneTemp(r25, Topt, T, omega=18.):
    '''
    leafmodel: temperature dependency
    Temperature dependency after June et al. 2004
    r25 : activity rate at 25 degC [unit depends on input]
    Topt: optimum temperature [degC]
    T: temperature [degC]
    omega: the difference in temperature from Topt at which activity rate falls
           to e-1 (0.37) of its value at Topt. (In the paper its on average
           18+-0.6 degC for Jmax)

    returns:
    r: activity rate at temperature T [unit depends on input]
    '''
    return r25 * np.ma.exp((-(T-Topt)**2 + (25.-Topt)**2)/(omega**2))

###############################################################################
def MedlynTemp(r25, E, delS, T, Hd):
    '''
    leafmodel: temperature dependency
    Temperature dependency after Medlyn et al. 2002
    r25 : activity rate at 25 degC [unit depends on input]
    E: activation energy [J]
    delS: entropy factor [J/mol/K]
    T: temperature [degC]
    Hd : decrease rate above optimum [J/mol]

    returns:
    r: activity rate at temperature T [unit depends on input]
    '''
    return r25 * np.ma.exp(((T-25.)*E)/(298.*R*(T+273.))) * \
                      (1. + np.ma.exp((298.*delS-Hd)/(298.*R))) / \
                      (1. + np.ma.exp(((T+273.)*delS-Hd)/((T+273.)*R)))

###############################################################################
def Topt(Hd, E, delS):
    '''
    leafmodel: temperature dependency
    Optimum temperature from entropy value after Medlyn et al. 2002
    Hd : decrease rate above optimum [J/mol]
    E: activation energy [J]
    delS: entropy factor [J/mol/K]

    returns:
    Topt: optimal temperature [degC]
    '''
    return Hd / (delS - R*np.ma.log(E/(Hd-E))) - 273.

###############################################################################
def sunfrac(G_of_theta, theta, lai):
    '''
    twoleafmodel: sunlit fraction of the canopy
    G_of_theta: leaf projection function at zenith angle theta [-]
    theta: zenith angle, 90. at horizon, 0. at noon [deg]
    lai: leaf area index [-]

    returns: sun lit fraction of lai [-]

    '''
    K = G_of_theta/np.ma.cos(np.deg2rad(theta))
    f_sun = (1.-np.ma.exp(-K*lai))/(K*lai)
    return f_sun

###############################################################################
def gb(u, d):
    '''
    leafmodel: leaf boundayr layer conductance
    Leaf boundary layer conductance after Bonan et al. 2002
    u: wind speed [m/s]
    d: leaf size [m]

    returns: leaf boundary layer conductance [m/s]
    '''
    return 1./(200.*np.ma.sqrt(d/np.ma.abs(u)))

###############################################################################
def ga(u, ustar):
    '''
    leafmodel: aerodynamic conductance
    Aerodynamic conductance
    u: wind speed [m/s]
    ustar: friction velocity [m/s]

    returns: aerodynamic conductance [m/s]
    '''
    return ustar**2/u

###############################################################################
def g2mol(g, Ta, P):
    '''
    leafmodel: conductance unit conversion
    Converts conductivity, e.g. aerodynamic or leaf boundary layer conductivity,
    from [m/s] to [mol/m2s]
    g: conductivity [m/s]
    Ta: air temperature [degC]
    P: air pressure [Pa]

    returns: conductivity [mol/m2s]
    '''
    return g / (R * (Ta+T0)/P)

###############################################################################
def leaftemp(Ta, H, ga, rho, cp):
    '''
    leafmodel: leaf temperature estimation
    Leaf temperature
    Ta: air temperature [degC]
    H: sensible heat flux [W/m2]
    ga: aerodynamic conductance [m/s]
    rho: air density [kg/m3]
    cp: heat capacity of dry air [J/kg/K]

    returns: leaf temperature [degC]
    '''
    return Ta+(H/(ga*rho*cp))


if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np
import networkx as nx
import random
import copy


__all__ = ['create_network', 'source_nodes', 'sink_nodes', 'plot_network']


def create_network(nnodes, nedges, maxdegree=None, maxoutdegree=None,
                   maxindegree=None, reverse=False, verbose=False):
    """
    This functions creates a directed acyclic graph with one source node, N nodes and M edges in total.
    Each node of the network is connected to the source node.
    Optionally the an upper degree of the nodes can be given to generate networks with are not too dense.
    The only exception from this upper bound is the source node.

    Definition
    ----------
    def create_network(nnodes, nedges, maxdegree=None)


    Input           Format                  Description
    -----           -----                   -----------
    nnodes          integer                 Number of nodes of the network. One node is the source not (no incoming edges, only outgoing).
                                            Multiple nodes might be sinks (no outgoing edges, only incoming).
    nedges          integer                 Number of (directed) edges of the network.
                                            Upper limit is ((maxdegree+1)*nnodes -  maxdegree - 1 ) / 2
                                               My calculations for that are:
                                               - source node has at most (nnodes-1) edges
                                               - all other (nnodes-1) nodes have at most maxdegree edges
                                               - all edges are now count twice because of directed property
                                            Lower limit is nnodes - 1
                                               Due to the fact that source node need to be connected at least to every other node
                                            TODO
                                               Redo this calculations when maxoutdegree and/or maxindegree are given
    maxdegree        integer                Upper degree of a node to otain not too dense networks. The only node is allowed to
                                            have a higher degree is the source node.
    maxoutdegree     integer                Upper limit of outgoing edges per node (except source node)
    maxindegree      integer                Upper limit of incoming edges per node (except source node)
    reverse          bool                   False: source node is the only source
                                            True:  after generatng the whole network (with one source node and maxdegree settings)
                                                   all edges are swapped such that one obtain a network with only one sink and multiple sources
                                                   Note: degree settings are made for one-source networks, i.e maxoutdegree -> maxindegree
                                                                                                               maxindegree  -> maxoutdegree
    verbose          bool                   If true print-outs, if false silent


    Description
    -----------
    (1) Adjacency matrix A[i,j]:
        - if A[i,j]=1 indicates that node i is connected with node j (i --> j)
        - if A[i,j]=0 indicates that connection i --> j does not exist
        A[i,j] != A[j,i] for directed graphs
    (2) Initialize A = 0
    (3) Fill only upper triangle of A (i < j) with nedges ones
        (this assures that graph is acyclic)
    (4) Diagonal needs to be zero such that there are no self-loops
    (5) Test if degree of all nodes (except first node = source node) have degree less eqaul max. degree given
    (6) Test if first node (source node) is connceted to all other nodes
        (assure that all nodes are connected and no subgraph exist)

    Restrictions
    ------------
    None

    Examples
    --------
    >>> random.seed('12345')
    >>> nnodes = 4
    >>> nedges = 5
    >>> G = create_network(nnodes, nedges, maxdegree=3)
    >>> nodes = G.nodes()
    >>> print('nodes: '+str(nodes))
    nodes: [0, 1, 2, 3]
    >>> edges = G.edges()
    >>> print('edges: '+str(edges))
    edges: [(0, 1), (0, 3), (1, 2), (1, 3), (2, 3)]
    >>> degrees = [ G.degree(ii) for ii in G.nodes() ]
    >>> print('degrees of nodes: '+str(degrees))
    degrees of nodes: [2, 3, 2, 3]

    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License.

    Copyright (c) 2016-2021 Juliane Mai

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Juliane Mai,    Oct 2016
    Modified, Matthias Cuntz, Sep 2021 - code refactoring
    """

    if maxdegree is None:
        maxdegree = nnodes - 1

    if maxoutdegree is None:
        maxoutdegree = nnodes - 1

    if maxindegree is None:
        maxindegree = nnodes - 1

    if (nedges < nnodes - 1):
        raise ValueError('Increase number of edges because no connected network is possible')

    if (nedges > int(((maxdegree+1)*nnodes - maxdegree - 1 ) / 2.0) ):
        raise ValueError('Decrease number of edges because at least one of the nodes will exceed upper limit of degrees')

    npos_edges = nnodes * (nnodes-1) / 2
    if (nedges > npos_edges):
        raise ValueError('Too many edges for given number of nodes! Upper limit is '+str(npos_edges))

    found_network = False
    while (not found_network):

        adjacency = np.zeros([nnodes, nnodes], dtype=int)
        cedge = 0
        while cedge < nedges:
            # integer random number of interval [0,npos_edges)
            rand = random.randrange(npos_edges)
            csum = np.cumsum(range(nnodes-1, 0, -1))

            # determine col and row for this number
            row = np.where(rand < csum)[0][0]
            if row == 0:
                sublist = np.arange(0, csum[row])
            else:
                sublist = np.arange(csum[row-1], csum[row])
            col = (row+1) + np.where(rand == sublist)[0][0]

            # check if this element is already set to "1",
            # if yes try another one
            if (adjacency[row, col] == 0):
                cedge += 1
                adjacency[row, col] = 1

        # calculate degrees of each node
        degrees     = [ np.sum(adjacency[:, ii]) + np.sum(adjacency[ii, :])
                        for ii in range(nnodes) ]
        degrees_in  = [ np.sum(adjacency[:, ii])
                        for ii in range(nnodes) ]
        degrees_out = [ np.sum(adjacency[ii, :])
                        for ii in range(nnodes) ]

        # print('degrees:     ',degrees)
        # print('degrees_in:  ',degrees_in)
        # print('degrees_out: ',degrees_out)

        # check if each node of network has degree less equal maxdegree
        degrees_satisified = ( all([ degrees[ii]     <= maxdegree
                                     for ii in range(1, nnodes) ]) and
                               all([ degrees_in[ii]  <= maxindegree
                                     for ii in range(1, nnodes) ]) and
                               all([ degrees_out[ii] <= maxoutdegree
                                     for ii in range(1, nnodes) ])
                               )

        if (degrees_satisified):
            if reverse:
                # swap all edges
                # only sink is last node
                adjacency = rotate(rotate(np.transpose(adjacency)))

            if verbose:
                print('degrees:     ', degrees)

            # create graph
            G = nx.MultiDiGraph()

            # add nodes to graph, IDs are 0...(n-1)
            G.add_nodes_from([ii for ii in range(nnodes)])
            nodes = G.nodes()
            if verbose:
                print('Added nodes: ', nodes)

            # add edges
            for ii in range(nnodes):
                for jj in range(ii+1, nnodes):
                    if adjacency[ii, jj] == 1:
                        G.add_edge(ii, jj)
            edges = G.edges()
            if verbose:
                print('Added edges: ', edges)

            if reverse:
                # check if all are connceted to sink node "nnodes-1"
                all_connected_to_0 = all([nx.has_path(G, ii, nnodes-1)
                                          for ii in range(1, nnodes)])
            else:
                # check if all are connceted to source node "0"
                all_connected_to_0 = all([nx.has_path(G, 0, ii)
                                          for ii in range(1, nnodes)])

            if all_connected_to_0:
                found_network = True
                # finished
            else:
                G.clear()
                if verbose:
                    print('New search required')
                    print('')
                # start new search
    return G


def source_nodes(G):
    """
    This functions returns all source nodes (no incoming edges) of a graph G.

    Definition
    ----------
    def source_nodes(G)


    Input           Format                  Description
    -----           -----                   -----------
    G               graph                   Networkx generated graph structure

    Description
    -----------
    None

    Restrictions
    ------------
    None

    Examples
    --------
    >>> random.seed('12345')
    >>> nnodes = 4
    >>> nedges = 5
    >>> G = create_network(nnodes, nedges, maxdegree=3)
    >>> sources = source_nodes(G)
    >>> print('source nodes: '+str(sources))
    source nodes: [0]

    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License.

    Copyright (c) 2016-2021 Juliane Mai

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Juliane Mai,    Oct 2016
    Modified, Matthias Cuntz, Sep 2021 - code refactoring
    """
    idx     = np.where( 0 == np.array([ G.in_degree(ii)
                                        for ii in G.nodes() ]))[0]
    sources = np.array(G.nodes())    # all nodes
    sources = sources[idx]           # only sources nodes
    return sources


def sink_nodes(G):
    """
    This functions returns all sink nodes (no outgoing edges) of a graph G.

    Definition
    ----------
    def sink_nodes(G)


    Input           Format                  Description
    -----           -----                   -----------
    G               graph                   Networkx generated graph structure

    Description
    -----------
    None

    Restrictions
    ------------
    None

    Examples
    --------
    >>> random.seed('12345')
    >>> nnodes = 4
    >>> nedges = 5
    >>> G = create_network(nnodes, nedges, maxdegree=3)
    >>> sinks = sink_nodes(G)
    >>> print('sink nodes: '+str(sinks))
    sink nodes: [3]

    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License.

    Copyright (c) 2016-2021 Juliane Mai

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Juliane Mai,    Oct 2016
    Modified, Matthias Cuntz, Sep 2021 - code refactoring
    """
    idx     = np.where( 0 == np.array([ G.out_degree(ii)
                                        for ii in G.nodes() ]))[0]
    sinks   = np.array(G.nodes())      # all nodes
    sinks   = sinks[idx]               # only sink nodes
    return sinks


def plot_network(G, fname_plot):
    """
    This functions plots a graph using either graphviz if available or standard
    matplotlib. Plots with graphviz are much more fancy since it reduces the
    number of intersections of plotted edges and draws round edges unstead of
    only straight lines.

    Definition
    ----------
    def plot_network(G, fname_plot)


    Input           Format                  Description
    -----           -----                   -----------
    G               graph                   Networkx generated graph structure
    fname_plot      string                  Filename of ouput pdf

    Description
    -----------
    None

    Restrictions
    ------------
    None

    Examples
    --------
    >>> random.seed('12345')
    >>> nnodes = 4
    >>> nedges = 5

    >>> G = create_network(nnodes, nedges, maxdegree=3)
    >>> fname_plot = 'test_dag.pdf'
    >>> plot_network(G, fname_plot)

    >>> import os
    >>> os.remove(fname_plot)

    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License.

    Copyright (c) 2016-2021 Juliane Mai

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Juliane Mai,    Oct 2016
    Modified, Matthias Cuntz, Sep 2021 - code refactoring
    """
    import matplotlib.pyplot as plt

    color   = [(0.237736, 0.340215, 0.575113),
               (0.404682, 0.548016, 0.244594),
               (0.786572, 0.753215, 0.297951)]
    sources = source_nodes(G)
    sinks   = sink_nodes(G)
    nodes = G.nodes()
    edges = G.edges()

    try:
        import pygraphviz as pgv

        A = pgv.AGraph(directed=True)
        A.add_edges_from(G.edges(),
                         color='grey', arrowhead='vee')

        alpha_fill  = 0.7
        alpha_bndr  = 0.0

        for ii in nodes:

            # determine color of node
            if ii in sinks:
                cc          = color[2]
                cc_hex_fill = "#%2x%2x%2x%2x" % (int(cc[0]*255),
                                                 int(cc[1]*255),
                                                 int(cc[2]*255),
                                                 int(alpha_fill*255))
                cc_hex_bndr = "#%2x%2x%2x%2x" % (int(cc[0]*255),
                                                 int(cc[1]*255),
                                                 int(cc[2]*255),
                                                 int(alpha_bndr*255))
            elif ii in sources:
                cc          = color[0]
                cc_hex_fill = "#%2x%2x%2x%2x" % (int(cc[0]*255),
                                                 int(cc[1]*255),
                                                 int(cc[2]*255),
                                                 int(alpha_fill*255))
                cc_hex_bndr = "#%2x%2x%2x%2x" % (int(cc[0]*255),
                                                 int(cc[1]*255),
                                                 int(cc[2]*255),
                                                 int(alpha_bndr*255))
            else:
                cc          = color[1]
                cc_hex_fill = "#%2x%2x%2x%2x" % (int(cc[0]*255),
                                                 int(cc[1]*255),
                                                 int(cc[2]*255),
                                                 int(alpha_fill*255))
                cc_hex_bndr = "#%2x%2x%2x%2x" % (int(cc[0]*255),
                                                 int(cc[1]*255),
                                                 int(cc[2]*255),
                                                 int(alpha_bndr*255))

            # plot node
            A.add_node(ii, color=cc_hex_bndr, fillcolor=cc_hex_fill,
                       style='filled')

        A.layout(prog='dot')
        A.draw(fname_plot)

    except ImportError:
        fig = plt.figure()

        pos   = nx.spring_layout(G)

        for ii in nodes:

            # determine color of node
            if ii in sinks:
                cc = color[2]
            elif ii in sources:
                cc = color[0]
            else:
                cc = color[1]

            # draw source node
            nx.draw_networkx_nodes(G, pos,
                                   nodelist=[ii],
                                   node_color=[cc],
                                   node_size=500,
                                   alpha=0.7)

        # draw edges
        nx.draw_networkx_edges(G, pos, alpha=0.7)

        # some math labels
        labels = {}
        for ii, inode in enumerate(nodes):
            labels[ii] = r'$'+str(inode)+'$'
        nx.draw_networkx_labels(G, pos, labels, font_size=16)

        plt.axis('off')
        fig.savefig(fname_plot)
        plt.close(fig)


def rotate(matrix):
    data = copy.deepcopy(matrix)
    for i in range(len(data)):
        for j in range(i+1, len(data[i])):
            temp = data[i][j]
            data[i][j] = data[j][i]
            data[j][i] = temp
    data = data[::-1]
    return data


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # nnodes    = 5
    # nedges    = 8
    # maxdegree = 3

    # # --------------------------
    # # create one network
    # # --------------------------
    # G = create_network(nnodes, nedges, maxdegree=maxdegree)
    # sink_nodes(G)
    # source_nodes(G)
    # plot_network(G, 'test_dag.pdf')

    # # --------------------------
    # # create 20 networks
    # # --------------------------
    # for inetwork in range(20):

    #     G = create_network(nnodes, nedges, maxdegree=maxdegree)

    #     # sources should be only "0" and should have at least one sink
    #     sources = source_nodes(G)
    #     sinks   = sink_nodes(G)

    #     # TODO add network to collection if not the same as an already existing: nx.is_isomorphic(G1, G2)

    #     # print some statistics
    #     print('number of nodes: ', G.number_of_nodes())
    #     print('number of edges: ', G.number_of_edges())
    #     print('source nodes:    ', sources)
    #     print('sink   nodes:    ', sinks)

    #     fname_plot = 'network_nodes_'+str(nnodes)+'_edges_'+str(nedges)+'_maxdeg_'+str(maxdegree)+'_'+str(inetwork+1)+'.pdf'
    #     plot_network(G,fname_plot)



#!/usr/bin/env python
import numpy as np
from scipy.spatial import Delaunay     # for triangulation
from scipy.integrate import dblquad      # area integral
from pyjams.jams.convex_hull import convex_hull  # convex hull of data points
from pyjams.jams.in_poly import in_poly      # test if point is in polygon
from pyjams.jams.area_poly import area_poly    # the area of a polygon


__all__ = ['volume_poly']


def volume_poly(func, x=None, y=None, tri=None, convexhull=False, area=False,
                allvol=False, **kwargs):
    """
    Volume of function above polygon. The polygon will be triangulated.
    Then the volume above each triangle is integrated, and alled summed up.
    If convexhull=True then the convex hull of the data points is taken
    as surface of the polygon.

    func must be func(y,x,*args) - note the y,x instead of x,y.
    This means that any unknown keyword in the call of volume_poly will
    be passed to func.


    Definition
    ----------
    def volume_poly(func, x=None, y=None, tri=None, convexhull=False,
                    area=False, allvol=False, **kwargs):


    Input
    -----
    func       function, func(y, x, *args)
               any unknown keyword will be passed to func


    Optional Input
    --------------
    Either
    x          1D array
        x coordinates
    y          1D array
        y coordinates
    or
    tri
        3D array (ntriangles, 3, 2), triangle coordinates (supercedes x, y)
                 (ntriangles, 3 corners, xy-coords for each corner)

    convexhull   bool
        integrate convex hull of polygon instead of polygon itself
        (ignored if tri is given)
    area         bool
        return also the area of the polygon (or the convex hull)
    allvol       bool
        return also all volumes and areas of triangles


    Output
    ------
    volume, integration_error
    polygon area if area
    triangle volumes, areas of triangles if allvol


    References
    ----------
    Juliane Mai's notes.


    Examples
    --------
    >>> def f1(y,x):
    ...     return 1
    >>> def f2(y,x,r):
    ...     return np.sqrt(r*r-x*x-y*y)
    >>> r = 1.0
    >>> n = 1000
    >>> np.random.seed(1)
    >>> x = 2.*np.random.random(n) - 1.
    >>> y = 2.*np.random.random(n) - 1.
    >>> ii = np.where(x*x+y*y <= r*r)[0]
    >>> x  = x[ii]
    >>> y  = y[ii]
    >>> from autostring import astr
    >>> print(astr(np.pi * r*r,3))
    3.142
    >>> v = volume_poly(f1, x, y, convexhull=True)
    >>> print(astr(v[0],3))
    2.996
    >>> v = volume_poly(f1, x, y, convexhull=True, area=True)
    >>> print(astr(v[0],3))
    2.996
    >>> print(astr(v[2],3))
    2.996

    >>> print(astr(0.5 * 4./3. * np.pi * r*r*r, 3))
    2.094
    >>> v = volume_poly(f2, x, y, r=r, convexhull=True)
    >>> print(astr(v[0], 3))
    2.072

    # Triangles outside
    >>> xy  = np.array(list(zip(x,y)))
    >>> d   = Delaunay(xy[:,:])
    >>> cxy = convex_hull(xy.transpose())
    >>> xs  = np.mean(cxy[:,0])
    >>> ys  = np.mean(cxy[:,1])

    # Construct triangles from convex hull vertices and centre of gravity
    >>> ntriangles = d.convex_hull.shape[0]
    >>> tri = np.empty((ntriangles,3,2), dtype=float)
    >>> for i in range(ntriangles):
    ...     tri[i,0,:] = xy[d.convex_hull[i,0],:]
    ...     tri[i,1,:] = xy[d.convex_hull[i,1],:]
    ...     tri[i,2,:] = [xs,ys]
    >>> v = volume_poly(f1, tri=tri, convexhull=True)
    >>> print(astr(v[0],3))
    2.996
    >>> v = volume_poly(f1, tri=tri, convexhull=True, area=True)
    >>> print(astr(v[0],3))
    2.996
    >>> print(astr(v[2],3))
    2.996
    >>> v = volume_poly(f2, tri=tri, r=r, convexhull=True)
    >>> print(astr(v[0], 3))
    2.072

    # Return all volumes and areas
    >>> v, e, vv, aa = volume_poly(f2, tri=tri, r=r, convexhull=True, allvol=True)
    >>> print(astr(vv[0], 3))
    0.056


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python
    library, Department of Computational Hydrosystems, Helmholtz Centre for
    Environmental Research - UFZ, Leipzig, Germany.

    Copyright (c) 2013 Matthias Cuntz, Juliane Mai - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.


    History
    -------
    Written,  Matthias Cuntz & Juliane Mai, Feb 2013
    Modified, Matthias Cuntz, Feb 2013 - tri
              Matthias Cuntz, Feb 2013 - ported to Python 3
              Matthias Cuntz, Sep 2021 - code refactoring
    """
    # Functions for the three lines of a triangle
    # Use the global variable tria
    def a(xx):  # line between x1 and x2
        if (tria[1, 0] - tria[0, 0]) == 0.:
            return b(xx)
        else:
            return ((tria[1, 1] - tria[0, 1]) / (tria[1, 0] - tria[0, 0]) *
                    (xx - tria[0, 0]) + tria[0, 1])

    def b(xx):   # line between x1 and x3
        return ((tria[2, 1] - tria[0, 1]) / (tria[2, 0] - tria[0, 0]) *
                (xx - tria[0, 0]) + tria[0, 1])

    def c(xx):   # line between x2 and x3
        if (tria[2, 0] - tria[1, 0]) == 0.:
            return b(xx)
        else:
            return ((tria[2, 1] - tria[1, 1]) / (tria[2, 0] - tria[1, 0]) *
                    (xx - tria[1, 0]) + tria[1, 1])

    # Case distinction if triangle points up or down
    def maxac(xx):
        return np.maximum(a(xx), c(xx))

    def minac(xx):
        return np.minimum(a(xx), c(xx))

    # Function to integrate
    def funcyx(yy, xx, shiftyx, *args):
        xx += shiftyx[1]
        yy += shiftyx[0]
        if args[0] == ():
            return func(yy, xx)
        else:
            kwargs = args[0]
            return func(yy, xx, **kwargs)

    if tri is not None:
        trigiven = True
        tri = np.array(tri)  # assure
        ntriangles = tri.shape[0]
        if (tri.shape[1] != 3) | (tri.shape[2] != 2):
            raise ValueError('volume_poly: tri must be triangles with dimension (ntriangles, 3, 2)')
    else:
        trigiven = False
        # check
        if np.size(x) != np.size(y):
            raise ValueError('volume_poly: x and y must have same dimensions')
        # Get convex hull and vertices
        xy  = np.array(list(zip(x, y)))
        d   = Delaunay(xy[:, :])
        cxy = convex_hull(xy.transpose())
        xs  = np.mean(cxy[:, 0])
        ys  = np.mean(cxy[:, 1])
        if convexhull:
            # Construct triangles from convex hull vertices and centre of gravity
            ntriangles = d.convex_hull.shape[0]
            tri = np.empty((ntriangles, 3, 2), dtype=float)
            for i in range(ntriangles):
                tri[i, 0, :] = xy[d.convex_hull[i, 0], :]
                tri[i, 1, :] = xy[d.convex_hull[i, 1], :]
                tri[i, 2, :] = [xs, ys]
        else:
            # All triangles
            tri = xy[d.vertices, :]
            ntriangles = tri.shape[0]

    flaeche = 0.
    tvol = 0.
    tvol_err = 0.
    volumes = np.zeros(ntriangles, dtype=float)
    areas   = np.zeros(ntriangles, dtype=float)
    # Calc mean semivariogramm over whole region
    thecount = 0
    for j in range(ntriangles):
        t    = tri[j, :, :]
        ii   = np.argsort(t[:, 0])
        tria = t[ii, :]
        xs   = np.mean(tria[:, 0])
        ys   = np.mean(tria[:, 1])
        # Select only Delaunay triangles that are inside the original polygon
        # i.e. exclude triangles in concave part of polygon
        # If convexhull=True then this is always true.
        check_in_poly = False
        if not trigiven:
            check_in_poly = in_poly([xs, ys], cxy[:, 0], cxy[:, 1]) >= 0
        if trigiven | check_in_poly:
            thecount  += 1
            areas[j]   = area_poly(tria[:, 0], tria[:, 1])
            flaeche   += areas[j]
            xmin       = np.amin(tria[:, 0])
            ymin       = np.amin(tria[:, 1])
            tria[:, 0] -= xmin # shift for integral>0
            tria[:, 1] -= ymin
            if (b(tria[1, 0]) > tria[1, 1]):  # triangle points down
                vol, err = dblquad(funcyx, tria[0, 0], tria[2, 0],
                                   maxac, b, args=([ymin, xmin], kwargs))
            else:                        # triangle points up
                vol, err = dblquad(funcyx, tria[0, 0], tria[2, 0],
                                   b, minac, args=([ymin, xmin], kwargs))
            volumes[j] = vol
            tvol      += vol
            tvol_err  += err * err
    tvol_err = np.sqrt(tvol_err)

    out  = [tvol]
    out += [tvol_err]
    if area:
        out += [flaeche]
    if allvol:
        out += [volumes[0:thecount]]
        out += [areas[0:thecount]]

    return out


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # def f1(y,x):
    #     return 1

    # def f2(y,x,r):
    #     return np.sqrt(r*r-x*x-y*y)

    # r = 1.0
    # n = 1000
    # np.random.seed(1)
    # x = 2.*np.random.random(n) - 1.
    # y = 2.*np.random.random(n) - 1.
    # ii = np.where(x*x+y*y <= r*r)[0]
    # x  = x[ii]
    # y  = y[ii]
    # print np.round(np.pi * r*r,3)
    # # 3.142
    # v = volume_poly(f1, x, y, convexhull=True)
    # print np.round(v[0],3)
    # # 2.996
    # v = volume_poly(f1, x, y, convexhull=True, area=True)
    # print np.round(v[0],3), np.round(v[2],3)
    # # 2.996 2.996

    # print np.round(0.5 * 4./3. * np.pi * r*r*r, 3)
    # # 2.094
    # v = volume_poly(f2, x, y, r=r, convexhull=True)
    # print np.round(v[0], 3)
    # # 2.072

    # # Triangles outside
    # xy  = np.array(zip(x,y))
    # d   = Delaunay(xy[:,:])
    # cxy = convex_hull(xy.transpose())
    # xs  = np.mean(cxy[:,0])
    # ys  = np.mean(cxy[:,1])
    # # Construct triangles from convex hull vertices and centre of gravity
    # ntriangles = d.convex_hull.shape[0]
    # tri = np.empty((ntriangles,3,2), dtype=float)
    # for i in range(ntriangles):
    #     tri[i,0,:] = xy[d.convex_hull[i,0],:]
    #     tri[i,1,:] = xy[d.convex_hull[i,1],:]
    #     tri[i,2,:] = [xs,ys]
    # v = volume_poly(f1, tri=tri, convexhull=True)
    # print np.round(v[0],3)
    # # 2.996
    # v = volume_poly(f1, tri=tri, convexhull=True, area=True)
    # print np.round(v[0],3), np.round(v[2],3)
    # # 2.996 2.996
    # v = volume_poly(f2, tri=tri, r=r, convexhull=True)
    # print np.round(v[0], 3)
    # # 2.072
    # v, e, vv, aa = volume_poly(f2, tri=tri, r=r, convexhull=True, allvol=True)
    # print np.round(vv[0], 3)
    # # 0.056




#!/usr/bin/env python
import numpy as np
import scipy.optimize as opt
import pyjams.functions as functions


def sigma_filter(x, y, z=3, func=functions.line_p, p=[0.,1.], plot=False, popt=False):

    """
        Checks a 1D-array for points deviating more than z standard deviations
        from a given function.
        The default function is a line and a 3-sigma filter is applied.
        If the data is very noisy then the filter profits from good starting parameters.


        Definition
        ----------
        def sigma_filter(x, y, z=3, func=functions.line_p, p=[0.,1.], plot=False):


        Input
        -----
        x            1D-array, or ND-array, which is filtered in the last dimension, i.e. x.shape[-1] == y.size
        y            1D-(masked) array


        Optional Input
        --------------
        z            multiplier to standard deviation used so that all deviations
                     abs(y-func(x,p)) > z*std(abs(y-func(x,p))) are masked
                     Default: 3
        func         func object. Default: functions.line_p
        p            initial function parameters, Default: [0,1]
        plot         True: plot iterative fits; Default: False
        popt         True: output also optimal parameters after filtering; False: output only mask
                     Default: False


        Output
        ------
        if popt:
            new_mask, popt    mask of y where y deviates more than z*std from fitted function, optimal parameters
        else:
            new_mask          mask of y where y deviates more than z*std from fitted function


        Examples
        --------
        >>> # Create some data
        >>> np.random.seed(1)
        >>> x = np.arange(40)
        >>> y = np.ma.array(2.+1.5*x+np.random.normal(0,1,40)) # line
        >>> y[2] = np.ma.masked
        >>> y[10] *= 10.
        >>> y[20] *= 10.
        >>> y[30] /= 10.
        >>> # detect outliers
        >>> new_mask = sigma_filter(x, y, z=5, plot=False)
        >>> # apply new mask to x
        >>> from autostring import astr
        >>> print(astr(np.ma.array(y, mask=new_mask), 1)[0:12])
        [' 3.6' ' 2.9' '--  ' ' 5.4' ' 8.9' ' 7.2' '12.7' '11.7' '14.3' '15.3' '--  ' '16.4']

        >>> def fun(x, p):
        ...     return(p[0]*np.sin(p[1]*x))
        >>> y = np.ma.array(fun(x, [2.,1.5])+np.random.normal(0,0.1,40)) # sine wave
        >>> y[2] = np.ma.masked
        >>> y[10] *= 10.
        >>> y[20] *= 10.
        >>> # detect outliers - line
        >>> print(sigma_filter(x, y, z=3, plot=False)[0:12])
        [False  True  True  True False  True False  True False False  True  True]
        >>> # detect outliers - sine
        >>> print(sigma_filter(x, y, z=5, p=[2.1,1.4], func=fun, plot=False)[0:12])
        [False False  True False False False False False False False  True False]

        >>> def fun2(x, p):
        ...     return(p[0]+p[1]*np.sin(p[2]*x))
        >>> y = np.ma.array(fun2(x, [2.,1.5,3.])+np.random.normal(0,0.1,40)) # sin wave
        >>> y[2] = np.ma.masked
        >>> y[10] *= 10.
        >>> y[20] *= 10.
        >>> # detect outliers
        >>> s, p = sigma_filter(x, y, z=5, p=[2.1,1.4,3.1], func=fun2, plot=False, popt=True)
        >>> print(s[0:12])
        [False False  True False False False False False False False  True False]
        >>> print(astr(p, 1))
        ['2.0' '1.5' '3.0']


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2018 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Feb 2014 - changed line_dev_mask
        Modified, MC, Dec 2014 - x.compressed error because x is non-masked
                  MC, May 2018 - pout, x ND-array
    """
    if plot:
        import matplotlib.pyplot as plt

    # create masked work array
    y_new  = np.ma.array(y.copy())
    if np.ma.count_masked(y_new) == 0: y_new.mask = np.zeros(y_new.shape)
    # Initialise fitted  parameters
    p_opt = np.copy(p)

    # loop until no outliers are detected anymore
    go_on = True
    while go_on:
        # fit function to data
        mm = np.where(~y_new.mask)[0]
        if x.ndim == 1:
            xx = x[mm]
        else:
            xx = x[...,mm]
        yy = y_new.compressed()
        p_opt = opt.fmin(functions.cost_abs, p_opt, args=(func,xx,yy), disp=False)

        # look for anomalies > z*standard dev
        dev  = np.abs(yy - func(xx,p_opt))
        sdev = np.std(dev, ddof=1)
        ii   = np.where(dev > (z*sdev))[0]

        if plot:
            fig = plt.figure('sigma_filter')
            sub = fig.add_subplot(111)
            sub.plot(x, y_new, 'ko-')
            iix = np.argsort(x)
            sub.plot(x[iix], func(x[iix],p_opt), 'b--')
            sub.plot(x[iix], func(x[iix],p_opt)-z*sdev, 'r--')
            sub.plot(x[iix], func(x[iix],p_opt)+z*sdev, 'r--')
            plt.show()

        if ii.size>0:
            y_new[mm[ii]] = np.ma.masked
        else:
            go_on = False

    # return mask where outliers are masked
    if popt:
        return y_new.mask, p_opt
    else:
        return y_new.mask


if __name__ == '__main__':
    import doctest
    doctest.testmod()

    # # Create some data
    # np.random.seed(1)
    # x = np.arange(40)
    # y = np.ma.array(2.+1.5*x+np.random.normal(0,1,40)) # line
    # y[2] = np.ma.masked
    # y[10] *= 10.
    # y[20] *= 10.
    # y[30] /= 10.
    # # detect outliers
    # new_mask = sigma_filter(x, y, z=5, plot=True)
    # # apply new mask to x
    # from autostring import astr
    # print(astr(np.ma.array(y, mask=new_mask), 1))
    # # [' 3.6' ' 2.9' '--  ' ' 5.4' ' 8.9' ' 7.2'
    # #  '12.7' '11.7' '14.3' '15.3' '--  ' '16.4'
    # #  '19.7' '21.1' '24.1' '23.4' '25.8' '26.6'
    # #  '29.0' '31.1' '--  ' '34.6' '35.9' '37.0'
    # #  '38.9' '38.8' '40.9' '41.6' '43.7' '46.0'
    # #  '--  ' '48.1' '49.3' '50.7' '52.3' '54.5'
    # #  '54.9' '57.7' '60.7' '61.2']

    # def fun(x, p):
    #     return(p[0]*np.sin(p[1]*x))
    # y = np.ma.array(fun(x, [2.,1.5])+np.random.normal(0,0.1,40)) # sine wave
    # y[2] = np.ma.masked
    # y[10] *= 10.
    # y[20] *= 10.
    # # detect outliers - line
    # print(sigma_filter(x, y, z=3, plot=True))
    # # [False  True  True  True False  True False  True False False  True  True
    # #   True False False  True  True False False False  True False False False
    # #   True False False False  True  True False False  True  True False False
    # #   True  True False False]
    # # detect outliers - sine
    # print(sigma_filter(x, y, z=5, p=[2.1,1.4], func=fun, plot=True))
    # # [False False  True False False False False False False False  True
    # #  False False False False False False False False False  True False
    # #  False False False False False False False False False False False
    # #  False False False False False False False]

    # def fun2(x, p):
    #     return(p[0]+p[1]*np.sin(p[2]*x))
    # y = np.ma.array(fun2(x, [2.,1.5,3.])+np.random.normal(0,0.1,40)) # sin wave
    # y[2] = np.ma.masked
    # y[10] *= 10.
    # y[20] *= 10.
    # # detect outliers
    # print(sigma_filter(x, y, z=5, p=[2.1,1.4,3.1], func=fun2, plot=True))
    # # [False False  True False False False False False False False  True False
    # #  False False False False False False False False  True False False False
    # #  False False False False False False False False False False False False
    # #  False False False False]



#!/usr/bin/env python
from random import randint, choice
from math import ceil, log
import os
import getpass
from pyjams.jams.sendmail import sendmail


__all__ = ['file_cipher', 'file_pass', 'set_up_cipher', 'wordEncrypt', 'wordDecrypt', 'sendfail']


# modified from http://code.activestate.com/recipes/577954-encrypt-and-decrypt-text-and-text-files-beta/
file_cipher = os.path.join(os.path.expanduser('~'),'.jams.cipher')
file_pass   = os.path.join(os.path.expanduser('~'),'.jams.email.cipher')
getVar = lambda searchList, ind:  [searchList[i] for i in ind]
find   = lambda searchList, elem: [[i for i, x in enumerate(searchList) if x == e] for e in elem]
mod    = lambda n, m:             n % m


# --------------------------------------------------------------------

def baseExpansion(n,c,b):
    '''baseExpansion - Power of ten transformation of keys'''
    i      = len(n)
    base10 = sum([pow(c,i-k-1)*n[k] for k in range(i)])
    j      = int(ceil(log(base10 + 1,b)))
    baseExpanded = [mod(base10//pow(b,j-p),b) for p in range(1,j+1)]
    return baseExpanded


# --------------------------------------------------------------------

def set_up_cipher():
    """
        Write or rewrite the cipher file: file_cipher

        This has to be done once before encryption and decryption can be used.


        Definition
        ----------
        def set_up_cipher():


        Output
        ------
        file with cipher in home directory


        Examples
        --------
        # Setup the encryption and decryption of password
        if not os.path.exists(cipher_file):
            set_up_cipher(cipher_file)

        # Store encrypted password
        pass_file = os.path.join(os.path.expanduser('~'),'.pass.cipher')
        if not os.path.exists(pass_file):
            print('')
            prompt = 'Enter password: '
            ipass = getpass.getpass(prompt)
            cpass = wordEncrypt(ipass)
            f = open(pass_file, 'w')
            f.write(' '.join([ str(i) for i in cpass ])+'\n')
            f.close()

        # Read encrypted password
        f = open(pass_file, 'r')
        cpass = f.readline()[:-1]
        f.close()
        password = wordDecrypt([ int(c) for c in cpass.split()])



        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                  http://code.activestate.com/recipes/577954-encrypt-and-decrypt-text-and-text-files-beta/
    """
    # key
    alpha = '1234567890qwertyuiop[]asdfghjkl;zxcvbnm,.!@#$%^&*()_+-=-{}:<>|QWERTYUIOPASDFGHJKLZXCVBNM ~`?'
    # cipher
    cipher = "".join([list(alpha)[randint(0,len(list(alpha))-1)] for i in range(5000)])

    if os.path.exists(file_cipher):
        os.remove(file_cipher)
    f = open(file_cipher, 'w')
    f.write(cipher)
    f.close()

# --------------------------------------------------------------------

def wordEncrypt(word):
    """
        Encrypt a word into list of keys using the cipher in file_cipher.


        Definition
        ----------
        def wordEncrypt(word):


        Input
        -----
        word     string


        Output
        ------
        list with numeric keys


        Examples
        --------
        # Setup the encryption and decryption of password
        if not os.path.exists(cipher_file):
            set_up_cipher(cipher_file)

        # Store encrypted password
        pass_file = os.path.join(os.path.expanduser('~'),'.pass.cipher')
        if not os.path.exists(pass_file):
            print('')
            prompt = 'Enter password: '
            ipass = getpass.getpass(prompt)
            cpass = wordEncrypt(ipass)
            f = open(pass_file, 'w')
            f.write(' '.join([ str(i) for i in cpass ])+'\n')
            f.close()

        # Read encrypted password
        f = open(pass_file, 'r')
        cpass = f.readline()[:-1]
        f.close()
        password = wordDecrypt([ int(c) for c in cpass.split()])


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                  http://code.activestate.com/recipes/577954-encrypt-and-decrypt-text-and-text-files-beta/
    """
    cipher = open(file_cipher).read()+'\n'
    cipherWord = find(cipher,list(word))
    
    keys = [randint(5001,7000), randint(2,5000)]

    encryptedWord = baseExpansion(list(map(choice, cipherWord)),keys[0],keys[1])
    encryptedWord.extend(keys)

    return list(map(int,encryptedWord))

# --------------------------------------------------------------------

def wordDecrypt(encryptedList):
    """
        Decrypt list of keys back into original word using the cipher in file_cipher.


        Definition
        ----------
        def wordDecrypt(encryptedList):


        Input
        -----
        encryptedList   list with numeric keys


        Output
        ------
        Original word (string)


        Examples
        --------
        # Setup the encryption and decryption of password
        if not os.path.exists(cipher_file):
            set_up_cipher(cipher_file)

        # Store encrypted password
        pass_file = os.path.join(os.path.expanduser('~'),'.pass.cipher')
        if not os.path.exists(pass_file):
            print('')
            prompt = 'Enter password: '
            ipass = getpass.getpass(prompt)
            cpass = wordEncrypt(ipass)
            f = open(pass_file, 'w')
            f.write(' '.join([ str(i) for i in cpass ])+'\n')
            f.close()

        # Read encrypted password
        f = open(pass_file, 'r')
        cpass = f.readline()[:-1]
        f.close()
        password = wordDecrypt([ int(c) for c in cpass.split()])


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                  http://code.activestate.com/recipes/577954-encrypt-and-decrypt-text-and-text-files-beta/
    """
    
    cipher        = open(file_cipher).read()+'\n'
    encryptedWord = encryptedList[0:len(encryptedList)-2]
    
    keys = encryptedList[len(encryptedWord):len(encryptedList)]
    
    decryptedList = map(int,baseExpansion(encryptedWord, keys[1], keys[0]))
    
    return "".join(getVar(cipher,decryptedList))


def sendfail(subject, message='', sender='jams.encrypt.sendfail'):
    """
        Send e-mail to e-mail addresses in file_pass, decrypting with file_cipher.


        Definition
        ----------
        sendfail(subject, message, sender='jams.encrypt.sendfail')


        Input
        -----
        subject   Subject of the e-mail
        message   Message body of e-mail


        Optional Input
        --------------
        sender      Sender name of e-mail


        Examples
        --------
        sendfail('This did not work.', sender='me@ufz.de')
        sendfail('This did not work.', 'Really.')


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
        Modified, MC, Oct 2015 - transfer to ufz library; message body
    """
    f = open(file_pass, 'r')
    email = f.readline()[:-1]
    cpass = f.readline()[:-1]
    f.close()
    user = getpass.getuser()
    password = wordDecrypt([ int(c) for c in cpass.split()])
    _ = sendmail(sender,
                 to       = email,
                 subject  = subject,
                 message  = message,
                 login    = user,
                 password = password)


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import numpy as np
from pyjams.fsread import fread
from pyjams.fsread import sread
from pyjams.jams.date2dec import date2dec
import time as t
import shutil as sh
import re
import os as os


################################################################################
def eddyspec(indir, cfile, hfile, rawfile, sltdir, tspfile='34_specmean.csv',
             cspfile='35_specmean.csv', hspfile='36_specmean.csv',
             novalue=-9999, plot=False):
    '''
    Provides plots with carbon and water fluxes from the rawfile together
    with the respective time lags in the cfile and hfile for the user to select
    'golden days' with sufficient flux and constant lags (about 6 half hour
    values during midday) to perform spectrum analysis with EddySpec and
    SpecMean (Kolle & Rebmann, 2007). When user selected desired time period,
    the respective *.slt files are printed to the console. The user can than
    use EddySpec and SpecMean with these files and the script continues after
    user is finished. Inductances for water and carbon are fitted and saved
    together with spectrum plots. It is recommended to repeat the spectrum
    analysis for different times within the year.


    Definition
    ----------
    eddyspec(indir, cfile, hfile, rawfile, sltdir, tspfile='34_specmean.csv',
             cspfile='35_specmean.csv', hspfile='36_specmean.csv', novalue=-9999):


    Input
    -----
    indir       str, path of the folder where results will be saved
    cfile       str, path of the carbon lag file
    hfile       str, path of the water lag file
    rawfile     str, path of the raw flux file
    sltdir      str, path of the folder containing the *.slt files (will contain
                EddySpec and SpecMean files after use)


    Optional Input
    --------------
    tspfile     str, name of the average temperature spectrum file
                (default: '34_specmean.csv')
    cspfile     str, name of the average carbon spectrum file
                (default: '35_specmean.csv')
    hspfile     str, name of the average water spectrum file
                (default: '36_specmean.csv')
    novalue     int, novalue in rawfile (default=-9999)


    Output
    ------
    (for each run in the year, a new folder is created in indir and EddySpec and
    SpecMean files are moved here from sltdir)
    c_specw.pdf    plot of the carbon spectrum
    h_specw.pdf    plot of the water spectrum
    spec_X_X.log   log file containing used *.slt files and inductances


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT
    License. The JAMS Python package originates from the former UFZ Python library,
    Department of Computational Hydrosystems, Helmholtz Centre for Environmental
    Research - UFZ, Leipzig, Germany.

    Copyright (c) 2014 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  AP, Jul 2014
    Modified, MC, Aug 2014 - clean up and Python 3
    '''
    ############################################################################
    # reading input files
    #lags
    lagdate   = np.array(sread('%s' %(cfile), nc=1, skip=1), dtype='|S16')
    day       = np.array([x[5:8] for x in lagdate.flatten()], dtype = '|S7').astype(float)
    hour      = np.array([x[8:10] for x in lagdate.flatten()], dtype = '|S2').astype(float)
    min       = np.array([x[10:12] for x in lagdate.flatten()], dtype = '|S2').astype(float)
    doysfloat = day + (hour + min/60.)/24.
    cl        = np.array(fread('%s' %(cfile), skip=1, cskip=1))
    hl        = np.array(fread('%s' %(hfile), skip=1, cskip=1))
    # fluxes
    fluxdate = date2dec(ascii = np.array(sread('%s' %(rawfile), nc=1, skip=1), dtype='|S16'))
    year     = np.array(sread('%s' %(rawfile), nc=1, skip=1), dtype='|S16')
    year     = np.array([x[6:10] for x in year.flatten()], dtype = '|S4').astype(int)
    fluxdate = fluxdate.flatten() - date2dec(yr = year, mo = 1, dy = 1, hr = 0, mi = 0, sc = 0)
    cf       = np.array(fread('%s' %(rawfile), skip=1, cskip=4, nc=1))
    cf       = np.where(cf == novalue, np.nan, cf).flatten()
    hf       = np.array(fread('%s' %(rawfile), skip=1, cskip=1, nc=1))
    hf       = np.where(hf == novalue, np.nan, hf).flatten()

    if plot:
        import matplotlib.pyplot as plt
        import matplotlib.backends.backend_pdf as pdf
        ############################################################################
        # plot lag c
        plt.figure(1)
        plt.plot(cl[:,0], cl[:,2], 'bo', label='minlag (sam)')
        plt.plot(cl[:,0], cl[:,5], 'ro', label='maxlag (sam)')
        plt.xlabel('DOY')
        plt.ylabel('Lags [Samples]')
        plt.title('c lags')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()

        ############################################################################
        # plot flux c
        plt.figure(2)
        plt.plot(fluxdate, cf, 'r-', label='cflux')
        plt.xlabel('DOY')
        plt.ylabel('cflux [?]')
        plt.title('c fluxes')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()

        ############################################################################
        # plot lag h
        plt.figure(3)
        plt.plot(hl[:,0], hl[:,2], 'bo', label='minlag (sam)')
        plt.plot(hl[:,0], hl[:,5], 'ro', label='maxlag (sam)')
        plt.xlabel('DOY')
        plt.ylabel('Lags [Samples]')
        plt.title('h lags')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()

        ############################################################################
        # plot flux h
        plt.figure(4)
        plt.plot(fluxdate, hf, 'b-', label='hflux')
        plt.xlabel('DOY')
        plt.ylabel('hflux [?]')
        plt.title('h fluxes')
        plt.axis('auto')
        plt.grid('on')
        plt.legend()
        plt.show()

    interval = []
    inp = True
    while inp:
        inpfrom = raw_input("From doy [ddd.ddd]: ")
        inpto = raw_input("To doy [ddd.ddd]: ")
        try:
            interval += [np.min(np.where(np.abs(doysfloat-float(inpfrom))<0.02083))]
            interval += [np.min(np.where(np.abs(doysfloat-float(inpto))<0.02083))]
            inp = False
        except (ValueError, IndexError):
            print('EddySpecWarning: type in floats(with . not ,), nothing else!')
            inp = True
    inpclag = raw_input("C lag: ")
    inphlag = raw_input("H lag: ")

    print(lagdate[interval[0]:interval[1]+1])
    print('\nDO EDDYSPEC AND SPECMEAN NOW!\n')
    inp = raw_input("Mean spectrums ready? [y/n]: ").lower()
    if inp == "y":
        pass
    else:
        import sys
        sys.exit()

    ############################################################################
    # move spectrum files from slt folder to spec folder
    if len(os.listdir(indir))==0:
        os.mkdir(indir+'/1')
        indir = indir+'/1'
    else:
        os.mkdir(indir+'/%i'%(int(os.listdir(indir)[-1])+1))
        indir = indir+'/'+os.listdir(indir)[-1]
    specpat  = re.compile('[0-9]*_specmean.csv')
    specpat2  = re.compile('[a-zA-Z0-9]*_[0-9]*_spec.csv')
    sltdirlist = os.listdir(sltdir)
    for file in sltdirlist:
        if bool(re.search(specpat, file)) | bool(re.search(specpat2, file)):
            sh.move('%s/%s' %(sltdir, file), indir)

    ############################################################################
    # conductance fitting
    # reading mean spec files
    tsp = np.array(fread('%s/%s' %(indir,tspfile), skip=1, nc=2))
    csp = np.array(fread('%s/%s' %(indir,cspfile), skip=1, nc=2))
    hsp = np.array(fread('%s/%s' %(indir,hspfile), skip=1, nc=2))

    # filter input specs for log10(input) != NaN
    global fcoc, fcoh
    fcoc = csp[np.where((np.invert(np.isnan(np.log10(tsp[:,1])))) & (np.invert(np.isnan(np.log10(csp[:,1])))))[0],1]
    fcoh = hsp[np.where((np.invert(np.isnan(np.log10(tsp[:,1])))) & (np.invert(np.isnan(np.log10(hsp[:,1])))))[0],1]
    xc = tsp[np.where((np.invert(np.isnan(np.log10(tsp[:,1])))) & (np.invert(np.isnan(np.log10(csp[:,1])))))[0],0]
    yc = tsp[np.where((np.invert(np.isnan(np.log10(tsp[:,1])))) & (np.invert(np.isnan(np.log10(csp[:,1])))))[0],1]
    xh = tsp[np.where((np.invert(np.isnan(np.log10(tsp[:,1])))) & (np.invert(np.isnan(np.log10(hsp[:,1])))))[0],0]
    yh = tsp[np.where((np.invert(np.isnan(np.log10(tsp[:,1])))) & (np.invert(np.isnan(np.log10(hsp[:,1])))))[0],1]

    # calculate deviations from t spec for different inductences and select the one with the smallest residual
    con = np.arange(0,5,0.01)
    devc = np.empty_like(con)
    devh = np.empty_like(con)

    for i in xrange(np.shape(con)[0]):
        devc[i] = np.nansum(np.log10(yc)-modc(xc,con[i]))
        devh[i] = np.nansum(np.log10(yh)-modh(xh,con[i]))

    conc = con[np.argmin(np.abs(devc))]
    conh = con[np.argmin(np.abs(devh))]

    if plot:
        ############################################################################
        # plot c spec
        fig1 = plt.figure(5)
        sub = fig1.add_subplot(111)
        sub.plot(np.log10(tsp[:,0]), np.log10(tsp[:,1]), 'ro-', label='T Spec')
        sub.plot(np.log10(csp[:,0]), np.log10(csp[:,1]), 'bo-', label='C Spec')
        sub.plot(np.log10(xc), modc(xc,conc),'go-', label='C Spec with %.2f Induc.' %(conc))
        plt.xlabel('F[Hz]')
        plt.ylabel('F[Hz]*Co(w,X)')
        sub.set_title('c Spectrum')
        sub.axis('auto')
        sub.grid('on')
        sub.legend(loc='best')

        ############################################################################
        # plot h spec
        fig2 = plt.figure(6)
        sub = fig2.add_subplot(111)
        sub.plot(np.log10(tsp[:,0]), np.log10(tsp[:,1]), 'ro-', label='T Spec')
        sub.plot(np.log10(hsp[:,0]), np.log10(hsp[:,1]), 'bo-', label='H Spec')
        sub.plot(np.log10(xh), modh(xh,conh),'go-', label='H Spec %.2f Induc.' %(conh))
        plt.xlabel('F[Hz]')
        plt.ylabel('F[Hz]*Co(w,X)')
        sub.set_title('h Spectrum')
        sub.axis('auto')
        sub.grid('on')
        sub.legend(loc='best')

        plt.show()

        ############################################################################
        # save figures
        pp1 = pdf.PdfPages('%s/c_spec%s.pdf' %(indir,rawfile[6:-4]))
        pp2 = pdf.PdfPages('%s/h_spec%s.pdf' %(indir,rawfile[6:-4]))
        fig1.savefig(pp1, format='pdf')
        fig2.savefig(pp2, format='pdf')
        pp1.close()
        pp2.close()

    ################################################################################
    # writing log file
    log = open('%s/spec_%i_%02i_%02i_%02i_%02i_%02i.log' %(indir,
                                                    t.localtime()[0], t.localtime()[1],
                                                    t.localtime()[2], t.localtime()[3],
                                                    t.localtime()[4], t.localtime()[5],)
                                                    , 'w')
    log.write('Inductence determination for the mean of:\n')
    for item in lagdate[interval[0]:interval[1]+1]:
        log.write('%s\n'%(item[0]))
    log.write('\nC lag: %s\n' %(inpclag))
    log.write('H lag: %s\n' %(inphlag))
    log.write('\nC Inductence: %.2f\n' %(conc))
    log.write('H Inductence: %.2f\n' %(conh))

# inductance fit functions for c
def modc(fc, conc):
    '''
    part of eddyspec: inductance model for c
    '''
    y = 1./(1. + 4. * np.pi**2 * fc**2 * conc**2)
    return np.log10(fcoc/y)

# inductance fit functions for h
def modh(fh, conh):
    '''
    part of eddyspec: inductance model for h
    '''
    y = 1./(1. + 4. * np.pi**2 * fh**2 * conh**2)
    return np.log10(fcoh/y)

if __name__ == '__main__':
    import doctest
    doctest.testmod()



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import numpy as np

def srrasa(xy, strata=5, n=3, plot=False):
    """
        Generates stratified random 2D points within a given rectangular area.


        Definition
        ----------
        def srrasa(xy, strata=5, n=3, plot=False):


        Input
        -----
        xy          list of floats (4), list with the x and y coordinates
                    enclosing the designated rectangle in the form [x1,x2,y1,y2]


        Optional Input
        --------------
        strata      int, number of strata per axis
        n           int, number of random points in each strata
        plot        bool, if True, stratas and points are plotted,
                    otherwise not


        Output
        ------
        rand_xy     ndarray (n,2), x and y coordinates of the stratified random
                    points in the given rectangular.


        Examples
        --------
        >>> # seed for reproducible results in doctest
        >>> np.random.seed(1)
        >>> # gives within the rectangle of the given coordinates
        >>> # 16 (4**2) stratas with 3 random points in each one.
        >>> rand_xy = srrasa([652219.,652290.,5772970.,5773040.], strata=4, n=3, plot=False)
        >>> from autostring import astr
        >>> print(astr(rand_xy[0:4,0:2],6,pp=True))
        [['6.522264e+05' '5.772975e+06']
         ['6.522318e+05' '5.772973e+06']
         ['6.522190e+05' '5.772972e+06']
         ['6.522401e+05' '5.772979e+06']]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2012-2013 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Nov 2012
        Modified, MC, Nov 2012 - default plot=False
                  AP, Dec 2012 - documentation change
                  MC, Feb 2013 - docstring
                  MC, Feb 2013 - ported to Python 3
    """

    # calculate strata steps
    sw = (xy[1]-xy[0])/strata
    sh = (xy[3]-xy[2])/strata
    xsteps = np.arange(xy[0],xy[1]+sw,sw)
    ysteps = np.arange(xy[2],xy[3]+sh,sh)

    # make output array
    rand_xy = np.empty((strata**2*n,2))

    # throw random points in each strata
    for j in range(strata):
        for i in range(strata):
            rand_xy[i*n+strata*n*j:(i+1)*n+strata*n*j,0] = (xsteps[i+1] - xsteps[i])*np.random.random(n) + xsteps[i]
            rand_xy[i*n+strata*n*j:(i+1)*n+strata*n*j,1] = (ysteps[j+1] - ysteps[j])*np.random.random(n) + ysteps[j]

    # plot stratas and random points within
    if plot:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        mpl.rc('font', size=20)
        mpl.rc('lines', linewidth=2)
        mpl.rc('axes', linewidth=1.5)
        mpl.rc('xtick.major', width=1.5)
        mpl.rc('ytick.major', width=1.5)
        mpl.rcParams['lines.markersize']=6

        fig = plt.figure('stratified random sampling')
        sub = fig.add_subplot(111, aspect='equal')
        sub.set_xlim(xy[0],xy[1])
        sub.set_ylim(xy[2],xy[3])
        for i in range(strata):
            sub.axhline(y=ysteps[i], color=(166/256., 206/256., 227/256.))
            sub.axvline(x=xsteps[i], color=(166/256., 206/256., 227/256.))
        sub.scatter(rand_xy[:,0],rand_xy[:,1],marker='+', s=60,
                    color=( 51/256., 160/256.,  44/256.))
        sub.set_xlabel('X')
        sub.set_ylabel('Y')
        sub.set_title('strata = %i, n = %i' %(strata,n))
        sub.xaxis.set_major_formatter(mpl.ticker.
                                      ScalarFormatter(useOffset=False))
        sub.yaxis.set_major_formatter(mpl.ticker.
                                      ScalarFormatter(useOffset=False))
        fig.autofmt_xdate(rotation=45)
        plt.tight_layout(pad=1, h_pad=0, w_pad=0)
        plt.show()

    return rand_xy



def srrasa_trans(xy,strata=5,n=3,num=3,rl=0.5,silent=True,plot=False):

    """
        Generates stratified random 2D transects within a given rectangular
        area.


        Definition
        ----------
        def srrasa(xy,strata=5,n=3,num=3,rl=0.5,silent=True,plot=False):


        Input
        -----
        xy          list of floats (4), list with the x and y coordinates
                    enclosing the designated rectangle in the form [x1,x2,y1,y2]


        Optional Input
        --------------
        strata      int, number of strata per axis
        n           int, number of random transects in each strata
        num         int, number of points in each transect
        rl          float [0. to 1.], relative length of transect with respect
                    to width of stratum
        silent      bool, if False, runtime diagnostics are printed to the
                    console, otherwise not
        plot        bool, if True, stratas and points are plotted,
                    otherwise not


        Output
        ------
        rand_xy     ndarray (n,2), x and y coordinates of the stratified random
                    transect points in the given rectangular.


        Examples
        --------
        >>> # seed for reproducible results in doctest
        >>> np.random.seed(1)
        >>> # gives within the rectangle of the given coordinates
        >>> # 16 (4**2) stratas with 3 random transects in each one.
        >>> # Each transect is 0.5*width_of_strata long and contains 5 points logarithmical distributed.
        >>> rand_xy = srrasa_trans([652219.,652290.,5772970.,5773040.], strata=4,
        ...                        n=3, num=5, rl=0.5, silent=True, plot=False)
        >>> from autostring import astr
        >>> print(astr(rand_xy[0:4,0:2],6,pp=True))
        [['6.522264e+05' '5.772983e+06']
         ['6.522276e+05' '5.772983e+06']
         ['6.522292e+05' '5.772983e+06']
         ['6.522315e+05' '5.772983e+06']]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2012-2013 Arndt Piayda, Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  AP, Nov 2012
        Modified, AP, Dec 2012 - documentation change
                  MC, Feb 2013 - ported to Python 3
    """

    # calculate strata steps
    sw = (xy[1]-xy[0])/strata
    sh = (xy[3]-xy[2])/strata
    xsteps = np.arange(xy[0],xy[1]+sw,sw)
    ysteps = np.arange(xy[2],xy[3]+sh,sh)
    tl = sw*rl

    # make output array
    rand_xy = np.empty((strata**2*n*num,2))

    o = 0
    for j in range(strata):
        for i in range(strata):
            for k in range(n):

                goon = True
                it = 0
                while goon:
                    # random seed in strata
                    seedx=(xsteps[i+1]-xsteps[i])*np.random.random(1)+xsteps[i]
                    seedy=(ysteps[j+1]-ysteps[j])*np.random.random(1)+ysteps[j]

                    # make logarithmic transect
                    tx   =np.arange(1,num+1)
                    dis  =np.sort(tl-np.log(tx)/np.max(np.log(tx))*tl)
                    seedx=np.repeat(seedx,num)+dis
                    seedy=np.repeat(seedy,num)

                    # random angle in strata [deg]
                    angle = 360 * np.random.random(1)

                    # rotate transect to random angle
                    seedx_trans = (-(seedy-seedy[0])*np.sin(np.deg2rad(angle))+
                                    (seedx-seedx[0])*np.cos(np.deg2rad(angle))+
                                     seedx[0])
                    seedy_trans =  ((seedy-seedy[0])*np.cos(np.deg2rad(angle))+
                                    (seedx-seedx[0])*np.sin(np.deg2rad(angle))+
                                     seedy[0])

                    # test if transect is in strata
                    if (((seedx_trans>xsteps[i]).all()) &
                        ((seedx_trans<xsteps[i+1]).all()) &
                        ((seedy_trans>ysteps[j]).all()) &
                        ((seedy_trans<ysteps[j+1]).all())):
                        goon = False

                    if not silent:
                        print('strata= (', i, ',', j, ')', ' it= ', it)
                    it += 1

                rand_xy[o:o+num,0] = seedx_trans
                rand_xy[o:o+num,1] = seedy_trans
                o += num

    # plot stratas and random transect points within
    if plot:
        import matplotlib as mpl
        import matplotlib.pyplot as plt
        mpl.rc('font', size=20)
        mpl.rc('lines', linewidth=2)
        mpl.rc('axes', linewidth=1.5)
        mpl.rc('xtick.major', width=1.5)
        mpl.rc('ytick.major', width=1.5)
        mpl.rcParams['lines.markersize']=6

        fig = plt.figure('stratified random transect sampling')
        sub = fig.add_subplot(111, aspect='equal')
        sub.set_xlim(xy[0],xy[1])
        sub.set_ylim(xy[2],xy[3])
        for i in range(strata):
            sub.axhline(y=ysteps[i], color=(166/256., 206/256., 227/256.))
            sub.axvline(x=xsteps[i], color=(166/256., 206/256., 227/256.))
        sub.scatter(rand_xy[:,0],rand_xy[:,1],marker='+', s=60,
                    color=( 51/256., 160/256.,  44/256.))
        sub.set_xlabel('X')
        sub.set_ylabel('Y')
        sub.set_title('strata = %i, n = %i, num = %i' %(strata,n,num))
        sub.xaxis.set_major_formatter(mpl.ticker.
                                      ScalarFormatter(useOffset=False))
        sub.yaxis.set_major_formatter(mpl.ticker.
                                      ScalarFormatter(useOffset=False))
        fig.autofmt_xdate(rotation=45)
        plt.tight_layout(pad=1, h_pad=0, w_pad=0)
        plt.show()

    return rand_xy


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
"""
    distributions.py        -> pdfs of additional distributions.
    sample_distributions.py -> sampling from additional distributions.


    Provided distributions
    ----------------------
    ep          Exponential Power of Box and Tiao (1992)
    exponential Exponential
    gauss       Gauss (Normal)
    laplace     Laplace
    multigauss  Multivariate Gauss (Normal) probability density function (pdf).
    multinorm   Multivariate normal (Gauss) probability density function (pdf).
    multinormal Multivariate normal (Gauss) probability density function (pdf).
    norm        Normal (Gauss)
    normal      Normal (Gauss)
    sep         Skew Exponential Power
    sep_fs      Skew Exponential Power after Fernandez & Steel (1998)
    st          Skew Student t
    st_fs       Skew Student t after Fernandez & Steel (1998)
    t           Student t


    Provided sampling from distributions
    ----------------------
    sample_ep      sample from exponential Power of Box and Tiao (1992)
    sample_sep     sample from skew exponential power distribution
    sample_sep_fs  sample from skew exponential power distribution after Fernandez and Steel (1998)
    sample_st      Sample from skew Student t
    sample_st_fs   Sample from skew Student t after Fernandez & Steel (1998)
    sample_t       Sample from Student t


    Examples
    --------
    >>> import numpy as np
    >>> print(str(laplace(0.)))
    0.5
    >>> print(str(laplace(0., 2., 2.) - 0.25/np.e))
    0.0


    License
    -------
    This file is part of the JAMS Python package, distributed under the MIT License.

    Copyright (c) 2016-2017 Matthias Cuntz, Juliane Mai, Dmitri Kavetski - mc (at) macu (dot) de

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.


    History
    -------
    Written,  MC,       May 2016
    Modified, JM+DK+MC, May 2016  - sampling from distributions
    Modified, MC,       Dec 2017  - multinormal
"""

from .distributions        import exponential, laplace
from .distributions        import gauss, normal, norm
from .distributions        import multigauss, multinormal, multinorm
from .distributions        import ep, sep, sep_fs, sep_fs_mean, sep_fs_std
from .distributions        import st, st_fs, st_fs_mean, st_fs_std, t

from .sample_distributions import sample_ep, sample_sep, sample_sep_fs
from .sample_distributions import sample_st, sample_st_fs, sample_t

# Information
__author__   = "Matthias Cuntz"
__version__  = '1.2'
__revision__ = "Revision: 20171222"
__date__     = 'Date: 22.12.2017'



#!/usr/bin/env python
from __future__ import division, absolute_import, print_function
import os

__all__ = ['tail']

def tail(ifile, n=10, noblank=False, comment='', keepnewline=False):
    """
        Return list with last n lines of file.
        Equivalent to Unix tail utility.

        If n is omitted, it defaults to 10.

        Blank (only whitespace) and comment lines can be excluded.


        Definition
        ----------
        def tail(ifile, n=10, noblank=False, comment='', keepnewline=False):


        Input
        -----
        ifile        file handle or filename


        Optional input parameters
        -------------------------
        n            Get the last n lines from file (default: 10)
        comment      Line gets excluded if first character of line is in comment sequence.
                     Sequence can be string, list or tuple.


        Options
        -------
        noblank      if True, excludes all lines that consists only of
                     whitespace characters.
        keepnewline  if True, newline character at end of string will not be removed.


        Output
        ------
        list with strings of last n lines in file


        Restrictions
        ------------
        For efficiency, the utility first seeks the file pointer to file.seek(-n*8192, os.SEEK_END)
        and then starts reading.
        If the lines are too long, 8192 might not be enough.
        The utility then attempts to determine the file seek from the length of the lines and
        redoes the reading. This is still experimental.


        Examples
        --------
        # Create some data
        >>> filename = 'test.dat'
        >>> ifile = open(filename,'w')
        >>> ifile.writelines('First line\\n')
        >>> ifile.writelines(' \\n')
        >>> ifile.writelines('# First comment\\n')
        >>> ifile.writelines('! Second comment\\n')
        >>> ifile.writelines('Last line\\n')
        >>> ifile.close()

        # Read last 10 lines
        >>> print(tail(filename))
        ['First line', ' ', '# First comment', '! Second comment', 'Last line']

        # Read last 4 lines
        >>> print(tail(filename, 4))
        [' ', '# First comment', '! Second comment', 'Last line']

        # Read last 4 lines, keep newlines in strings
        >>> print(tail(filename, n=4, keepnewline=True))
        [' \\n', '# First comment\\n', '! Second comment\\n', 'Last line\\n']

        # Read last 4 lines excluding blank lines
        >>> print(tail(filename, 4, noblank=True))
        ['First line', '# First comment', '! Second comment', 'Last line']

        # Read last 4 lines excluding lines starting with #
        >>> print(tail(filename, 4, comment='#'))
        ['First line', ' ', '! Second comment', 'Last line']

        # Read last 4 lines excluding lines starting with # and !
        >>> print(tail(filename, 4, comment='#!'))
        ['First line', ' ', 'Last line']

        # Read last 4 lines excluding lines starting with # and S
        >>> print(tail(filename, 4, comment='#S'))
        ['First line', ' ', '! Second comment', 'Last line']

        # Read last 4 lines excluding lines starting with # and L
        >>> print(tail(filename, 4, comment=('#','L')))
        ['First line', ' ', '! Second comment']

        # Read last 4 lines excluding lines starting with # and ! but comment given as list not string
        >>> print(tail(filename, 4, comment=['#','!']))
        ['First line', ' ', 'Last line']

        # Read last 4 lines excluding lines starting with # and ! and blank lines
        >>> print(tail(filename, 4, comment='#!', noblank=True))
        ['First line', 'Last line']

        # Read last 2 lines
        >>> print(tail(filename, n=2))
        ['! Second comment', 'Last line']

        # Read last line excluding lines starting with # and ! and blank lines
        >>> print(tail(filename, 1, comment='#!', noblank=True))
        ['Last line']

        # Read last lines from an open file handle rather than a file name
        >>> f = open(filename, 'r')
        >>> print(tail(f, 1, comment='#!', noblank=True))
        ['Last line']
        >>> f.close()

        # Check intermediate line lengths
        >>> n=1000
        >>> f = open(filename, 'w')
        >>> print('a'*n, file=f)
        >>> print('b'*n, file=f)
        >>> print('c'*n, file=f)
        >>> print('d'*n, file=f)
        >>> f.close()
        >>> ll = tail(filename, n=2)
        >>> if (ll[0] != 'c'*n) or (ll[1] != 'd'*n):
        ...     print('failed')

        # Check long line lengths
        >>> n=10000
        >>> f = open(filename, 'w')
        >>> print('a'*n, file=f)
        >>> print('b'*n, file=f)
        >>> print('c'*n, file=f)
        >>> print('d'*n, file=f)
        >>> f.close()
        >>> ll = tail(filename, n=2)
        >>> if (ll[0] != 'c'*n) or (ll[1] != 'd'*n):
        ...     print('failed')

        # Clean up
        >>> import os
        >>> os.remove(filename)


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
                  MC, Nov 2016 - adapted file handling to Python 2 and 3
    """
    import sys
    if sys.version_info > (3,0):
        import io
        f = ifile if isinstance(ifile,io.TextIOWrapper) else open(ifile,"r")
    else:
        f = ifile if isinstance(ifile,file) else open(ifile,"r")

    # Seek
    bufsize = 8192*n
    try:    # might be out of file
        f.seek(-bufsize, os.SEEK_END)
    except: # stay were you are
        pass

    # Read lines
    doit = True
    while doit:
        count = 0
        liste = [[]]*n
        imax  = []
        for line in f:
            if not line:
                break
            if noblank and not line.strip():
                continue
            if line[0] in comment:
                continue
            if not keepnewline:
                line = line.strip("\n")
            liste.insert(n, line)
            tmp = liste.pop(0)
            imax.append(len(line.strip("\n"))) # lines include \0 at the end.
            count += 1
        imax = max(imax)
        if (imax < bufsize//n) or (count > n):
            doit = False
        else:
            bufsize = imax*8*n # 8 is for safety
            try:    # might be out of file
                f.seek(-bufsize, os.SEEK_END)
            except:
                f.seek(0, os.SEEK_SET)

    if f is not ifile:
        f.close()

    if count < n:
        return liste[-count:]
    else:
        return liste


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)

    # filename = 'test.dat'
        
    # n=1000
    # f = open(filename, 'w')
    # print('a'*n, file=f)
    # print('b'*n, file=f)
    # print('c'*n, file=f)
    # print('d'*n, file=f)
    # f.close()
    # ll = tail(filename, n=2)
    # if (ll[0] != 'c'*n) or (ll[1] != 'd'*n):
    #     print('failed')

    # n=10000
    # f = open(filename, 'w')
    # print('a'*n, file=f)
    # print('b'*n, file=f)
    # print('c'*n, file=f)
    # print('d'*n, file=f)
    # f.close()
    # ll = tail(filename, n=2)
    # if (ll[0] != 'c'*n) or (ll[1] != 'd'*n):
    #     print('failed')

    # import os
    # os.remove(filename)
